# TODO: should this go in nems.layers directory instead?
#       or maybe nems.tools?

import logging
import inspect
import re

log = logging.getLogger(__name__)


class KeywordMissingError(Exception):
    """Raised when a keyword lookup fails."""
    def __init__(self, value):
        default_message = f'"{value}" not found in keyword registry.'
        super().__init__(default_message)


class Keyword:
    """Maps a Layer to a string representation.
    
    See also
    --------
    For additional information on keyword formatting:
    nems.layers.base.Layer.from_keyword

    """

    def __init__(self, key, parse, source_string=None):
        """
        Parameters
        ----------
        key : str
            A short-hand description of a Layer with a particular set of
            options.
        parse : callable
            Function that parses the full keyword string. In most cases this
            should be the `from_keyword` method of the relevant Layer.
        source_string : str, optional
            Location (i.e., python layer or file) where `parse` was defined.

        Example
        -------
        `key = wc.g` would map to WeightChannels(parameterization='gaussian')

        """
        self.key = key
        self.parse = parse
        self.source_string = source_string if source_string is not None else ''

    def __repr__(self):
        return str(self.parse(self.key))



class KeywordRegistry:
    '''A collection of Keywords registered by the `@layer` decorator.

    A `KeywordRegistry` behaves similar to a dictionary, except that
    `KeywordRegistry[key_string]` will trigger a function call with `key_string`
    as the only argument. The function to call is specified by the Keyword
    instance referenced by the leading portion of the key: either before the
    first period or until the first non-alpha character.
    
    Example
    -------
        def simple_keyword(kw):
            return kw.split('.')
        kws = KeywordRegistry(name='test')
        kws['mykey'] = simple_keyword

        In[0]: kws['mykey.test']
        Out[0]: ['mykey', 'test']

    '''

    def __init__(self, name='registry', warn_on_overwrite=True,
                 set_obj_name=False):
        """
        Parameters
        ----------
        name : str
            Just a name to distinguish the registry's purpose, incase multiple
            registries are defined.
        warn_on_overwrite : bool
            If True, raise a RuntimeWarning when a keyword is overwritten to
            help debug name clashes.
        set_obj_name : bool
            If true, set `obj.name = kw_head` during indexing.

        """
        self.keywords = {}
        self.list = []
        self.name = name
        self.warn_on_overwrite = warn_on_overwrite
        self.set_obj_name = set_obj_name

    def __getitem__(self, kw_string):
        kw = self.lookup(kw_string)
        obj = kw.parse(kw_string)

        if (self.set_obj_name):
            # Type check is done this way to avoid circular import issues.
            inheritance_list = inspect.getmro(type(obj))
            checks = [cls.__name__.split('.')[-1] != 'Layer'
                      for cls in inheritance_list]
            if all(checks):
                raise TypeError(
                    '@layer functions must return a Layer instance.'
                    )
            if obj._name is None:
                kw_head = kw_string.split('.')[0]
                obj._name = kw_head

        return obj

    def __setitem__(self, kw_head, parse, source=None):
        if kw_head in self.keywords and self.warn_on_overwrite:
            raise RuntimeWarning(
                f'Keyword: {kw_head} overwritten in registry: {self.name}'
                )
        self.keywords[kw_head] = Keyword(kw_head, parse, source_string=source)
        self.list.append(f'{kw_head}:    {source}')

    def kw_head(self, kw_string):
        """Identifies the leading portion of a keyword string.
        
        Parameters
        ----------
        kw_string : str

        Returns
        -------
        kw_head : str
            Substring of `kw_string` that comes before the first period. If no
            period is present, this will be the first substring of  alpha-only
            characters instead.
        
        """

        # if the full kw_string is in the registry as-is, then it's a
        # backwards-compatibility alias and overrides the normal kw head rule.
        if kw_string in self.keywords:
            return kw_string
        # Look for '.' first. If not present, use first alpha-only string
        # as head instead.
        h = kw_string.split('.')
        if len(h) == 1:
            # no period, do regex for first all-alpha string
            alpha = re.compile('^[a-zA-Z]*')
            kw_head = re.match(alpha, kw_string).group(0)
        else:
            kw_head = h[0]
        return kw_head

    def lookup(self, kw_string):
        """Find `kw_string` in the registry using the leading portion.
        
        See also
        --------
        `KeywordRegistry.kw_head`
        
        """
        kw_head = self.kw_head(kw_string)
        if kw_head not in self.keywords:
            raise KeywordMissingError(kw_head)

        return self.keywords[kw_head]

    def source(self, kw_string):
        """Fetch the definition location for `kw_string`."""
        return self.lookup(kw_string).source_string

    def register_function(self, obj, name=None, obj_type=None):
        """Add function `obj` to the registry.
        
        Parameters
        ----------
        obj : callable
        name : str; optional.
            Name of KeyWord to store in registry. Defaults to `obj.__name__`.
        obj_type : str; optional.
            If specified, raise TypeError if `type(obj).__name__` does not match
            `obj_type`. String specification is used to avoid circular imports
            for decorators.

        """

        if name is None:
            name = obj.__name__
        log.debug("%s lib registering function: %s", self.name, name)
        
        try:
            # Default to module rather than file, to maintain portability.
            location = str(obj.__module__) + "." + obj.__name__
        except AttributeError:
            location = str(obj.__name__)
            
        self.keywords[name] = Keyword(name, obj, location)
        self.list.append(f'{name}:    {location}')

    def __iter__(self):
        """Iterate over registered keywords."""
        return self.keywords.__iter__()

    def __next__(self):
        """Iterate over registered keywords."""
        return self.keywords.__next__()

    def info(self, kw=None):
        """Look up sources of registered keyword(s).

        Parameters
        ----------
        kw : str or None

        Returns
        -------
        dict
            Contains source information for keywords containing string `kw`,
            or all keywords if `kw` is None.
    
        """
        kw_list = list(self.keywords.keys())
        s = {}
        for k in kw_list:
            if (kw is None) or kw in k:
                s[k] = self.source(k)
        return s


# Create registries here so they can be updated as new layers are imported.
keyword_lib = KeywordRegistry(
    name="layers", warn_on_overwrite=True, set_obj_name=True
    )

def layer(name=None):
    """Decorator for `Layer.from_keyword()` methods.
    
    Returns
    -------
    decorator : function
        Registers a decorated function in `keyword_lib`.

    Examples
    --------
    >>> @layer('cool')
    >>> def from_keyword(keyword):
    ...     # parse keyword into some Layer options
    ...     option1, option2 = my_kw_parser(keyword)
    ...     return CoolLayer(option1, option2)
    
    """
    def decorator(func):
        # If func has an underlying __func__, use that instead to avoid
        # error from staticmethod not having __module__ or __name__ attributes.
        func = getattr(func, '__func__', func)

        # Raise error if func doesn't accept any arguments.
        try:
            args = inspect.getfullargspec(func)
            _ = args[0]
        except TypeError:
            raise TypeError(
                f'Keyword {name} must accept a keyword string '
                'as its first argument.'
            )

        # If there aren't any syntax errors, register the keyword.
        keyword_lib.register_function(func, name=name, obj_type='Layer')
        return func

    return decorator



"""A collection of preprocessing utilities.

# TODO: These contents don't all exist yet, sketching this out based on
#       the strfy proposal document.
Contents:
    `spectrogram`: Convert sound waveforms to spectrograms.
    `filters.py`: Transform data using smoothing, high-pass filter, etc.
    `normalization.py`: Scale data to a standard dynamic range.
    `raster.py`: Convert data to continuous time series (ex: spikes -> PSTH).
    `split.py`: Separate data into estimation & validation sets.
    `mask.py`: Exclude subsets of data based on various criteria.
    `merge.py`: Align and combine datasets for fitting a single large model.

"""

from .spectrogram import spectrogram, fft_gammagram, gammagram
from .normalization import minmax, undo_minmax, joint_minmax
from .raster import raster_to_spike_times, spike_times_to_raster
from .split import (indices_by_fraction, split_at_indices,
                    get_jackknife_indices, get_jackknife)



"""Tools for visualizing the inputs, outputs, and parameters of Layers."""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from nems.layers.stp import ShortTermPlasticity
from nems.preprocessing.normalization import minmax, joint_minmax
from nems.distributions import Normal, HalfNormal
from .tools import standardize_axes


def stp_test_input():
    """Generate synthetic data to probe STP's input/output relationship.

    Each channel is structured as:
        Row 0: Random [0,1)
        Row 1: Square-wave
        Row 2: All zero

    Returns
    -------
    np.ndarray
        Shape (100, 3).
    
    """
    input = np.random.rand(100, 3)
    input[:, 1] = 0
    input[10:20, 1] = 1
    input[25:35, 1] = 1
    input[60:90, 1] = 1
    input[:, 2] = 0

    return input


def stp_input_output(n=5, quick_eval=False, figsize=None):
    """Visualize STP input/output relationship for random parameter values.
    
    Creates a 3x3 grid of input/output plots.
    Each column is associated with one parameter sample.
        Column 0: Default parameters.
        Column >= 1: Randomly sampled (*not* `stp.sample_from_priors`.)
                     u:       Normal(mean=0.25, sd=0.25)  s.t.   |u| >= 0.001
                     tau: HalfNormal(           sd=5   )  s.t. |tau| >= 0.100
        For cases where u < 0 (facilitation), input & output will be normalized
        such 
    Each row is associated with one channel of the synthetic input data:
        Row 0: Random [0,1)
        Row 1: Square-wave
        Row 2: All zero.

    Parameters
    ----------
    n : int; default=5.
        Number of random parameter sets to sample. One column is plotted for
        each set, so much larger or smaller values of `n` will likely require
        a change to `figsize`.
    quick_eval : bool; default=False.
        Determines which evaluation algorithm is used (see `nems.layers.stp`).
    figsize : tuple of int; optional.
        (width, height) in inches. Default is ((n*3)+1, 8).
    
    Returns
    -------
    matplotlib.pyplot.Figure
    
    """

    # Generate synthetic input
    # 0: random, 1: square, 2: zero
    input = stp_test_input()

    if figsize is None: figsize = ((n*3)+1, 8)
    fig, axes = plt.subplots(3, n, sharey='col', figsize=figsize)
    if n == 1: axes = axes[..., np.newaxis]  # so that subsequent indexing works
    titles = ['Random [0,1)', 'Square', 'Zeros']

    # Generate output data for n parameter sets and add it to the figure.
    for i in range(n):
        stp = ShortTermPlasticity(shape=(3,), quick_eval=quick_eval)

        if i != 0:
            # Make parameter values random, but matched for all channels.
            u, tau = _stp_visualization_parameters(quick_eval=quick_eval)
            stp.set_parameter_values(
                u=[u for _ in range(3)],
                tau=[tau for _ in range(3)]
                )
        else:
            # Use default values, don't resample.
            u = stp.get_parameter_values('u')[0][0]
            tau = stp.get_parameter_values('tau')[0][0]

        # Get Layer output
        output = stp.evaluate(input)

        # Normalize between 0 and 1 for easier visualization
        # (already true for depression, but not facilitation)
        this_input, this_output = joint_minmax(input, output)

        # Plot input and output for each channel
        for j, ax in enumerate(axes[:,i]):
            ax.plot( this_input[:, j], color='black', label='Input')
            ax.plot(this_output[:, j], color='slateblue', label='Output')
            ax.xaxis.set_visible(False)
            ax.set_title(titles[j])
            standardize_axes(ax)

        # Label parameter values
        ax.text(95, 0.5, f'u: {np.round(u,3)}\ntau: {np.round(tau,3)}',
                ha='right')
        ax.legend(frameon=False)
        ax.xaxis.set_visible(True)
        ax.set_xlabel('Time (bins)')
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    fig.suptitle('STP input/output')
    fig.tight_layout()

    return fig


def _stp_visualization_parameters(quick_eval=False):
    """Sample u and tau. Internal for `stp_input_output`."""
    # Manually sample parameters.
    if not quick_eval:
        u = Normal(mean=0.25, sd=0.25).sample(1)
    else:
        u = HalfNormal(sd=0.1).sample(1)
    tau = HalfNormal(5).sample(1)
    if u < 0: u *= 0.1
    if np.abs(u) < 1e-3: u = np.sign(u)*1e-3
    if tau < 0.1: tau = 0.1

    return u, tau



"""Collection of sampling Distributions.

Intended to serve as priors for  `nems.layers.base.Parameter`.
Modified from code by Dr. Brad Buran (github: bburan).

"""

from .normal import Normal
from .half_normal import HalfNormal
from .uniform import Uniform
from .beta import Beta
from .gamma import Gamma
from .exponential import Exponential



# This code is derived from the gammatone toolkit, licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING

"""Represent waveforms with spectrograms using the Fast Fourier Transform."""

from __future__ import division
import numpy as np

from .filters import erb_space, make_erb_filters
from .gammatone import gtgram_strides


def spectrogram(wave, fft_size=None, fs=44000, window_time=0.01,
                hop_time=0.01):
    """Compute a simple spectrogram for sound `wave`.

    Parameters
    ----------
    wave : np.ndarray.
        Sound waveform with shape (T,) or (T,1).
    fft_size : int, divisible by 2; optional.
        Number of points for the Discrete Fourier Transform.
        Default is `2^ceil(log_2(2*window_time*fs))`.
        (`n` in Dan Ellis `specgram`).
    fs : int; default=44000.
        Sampling frequency. `gtgram_strides` uses this to scale `window_time`
        and `hop_time` to convert them to units of bins.
    window_time : float; default=0.01.
        Length of integration window. Default of 0.01 corresponods to 1/100hz,
        where 100hz is the most common spike raster sampling rate used by LBHB.
    hop_time : float; default=0.01.
        Stepsize of window advancement for successive columns. Default of 0.01
        corresponods to 1/100hz, where 100hz is the most common spike raster
        sampling rate used by LBHB.

    Returns
    -------
    np.ndarray, dtype=complex.
        Shape (Tau, fft_size/2 + 1)
        Where `Tau = 1 + floor((sound_length - fft_size)/hop_bins))`.
        If the default value is used for `fft_size`, this is equivalent to
        downsampling the spectrogram to a sampling rate of `1/hop_time`, with
        some rounding error to get integer bins.

    Copyright
    ---------
    Based on Dan Ellis' myspecgram.m,v 1.1 2002/08/04.

    """

    if hop_time <= 0: raise ValueError("Must have a hop size greater than 0")
    if fft_size is None:
        fft_size = _default_fft_size(window_time, fs)

    sound_length = wave.shape[0]
    window_bins, hop_bins, _ = gtgram_strides(fs, window_time, hop_time, 0)
    win = _spectrogram_window(fft_size, window_bins)

    t = 0

    # pre-allocate output array
    n_time_bins = 1 + int(np.floor((sound_length - fft_size)/hop_bins))
    spectrogram = np.zeros((n_time_bins, (1 + fft_size//2)), np.dtype(complex))

    for b in range(0, sound_length - fft_size, hop_bins):
      u = win * wave[b : b + fft_size]
      fft = np.fft.fft(u)
      spectrogram[t, :] = fft[0 : (1 + fft_size//2)]
      t = t + 1

    return spectrogram


def _spectrogram_window(nfft, nwin):
    """Window calculation used in specgram replacement function.
    
    Hann window of width `nwin` centred in an array of width `nfft`.
    Internal for `spectrogram`.

    """
    halflen = nwin // 2
    halff = nfft // 2 # midpoint of win
    acthalflen = int(np.floor(min(halff, halflen)))
    halfwin = 0.5 * ( 1 + np.cos(np.pi * np.arange(0, halflen+1)/halflen))
    win = np.zeros((nfft,))
    win[halff : halff + acthalflen] = halfwin[0 : acthalflen]
    win[halff : halff - acthalflen : -1] = halfwin[0 : acthalflen]

    return win

def _default_fft_size(window_time, fs):
    """Get default number of points for DFT, internal for `spectrogram`."""
    return int(2 ** (np.ceil(np.log2(2 * window_time * fs))))


def fft_gammagram(wave, fs=44000, window_time=0.01, hop_time=0.01,
                  channels=18, f_min=200.0, f_max=None):
    """Approximate a gammatone filter spectrogram using FFT.

    A matrix of weightings is calculated using `fft_weights`, and applied to
    the FFT of the input signal, `wave`, using sample rate `fs`. The result is
    an approximation of full filtering using an ERB gammatone filterbank
    (as per `gammagram`).

    `f_min` determines the frequency cutoff for the corresponding gammatone
    filterbank. `window_time` and `hop_time` (both in seconds) are the size
    and overlap of the spectrogram columns.

    Parameters
    ----------
    wave : np.ndarray.
        Sound waveform with shape (T,) or (T,1).
    fs : int; default=44000.
        Sampling frequency. `gtgram_strides` uses this to scale `window_time`
        and `hop_time` to convert them to units of bins.
    window_time : float; default=0.01.
        Length of integration window. Default of 0.01 corresponods to 1/100hz,
        where 100hz is the most common spike raster sampling rate used by LBHB.
    hop_time : float; default=0.01.
        Stepsize of window advancement for successive columns. Default of 0.01
        corresponods to 1/100hz, where 100hz is the most common spike raster
        sampling rate used by LBHB.
    channels : int.
        Number of frequency channels in the spectrogram-like output.
    f_min : float; default=200.0.
        Lower frequency cutoff.
    f_max : float; optional.
        Upper frequency cutoff. If not specified, will be set to `fs/2`.

    Returns
    -------
    np.ndarray
        With shape (Tau, `channels`)
        Where `Tau = 1 + floor((sound_length - fft_size)/hop_bins))`.
        This is equivalent to downsampling the spectrogram to a sampling rate
        of `1/hop_time`, with some rounding error to get integer bins.

    Copyright
    ---------
    2009-02-23 Dan Ellis dpwe@ee.columbia.edu
    (c) 2013 Jason Heeris (Python implementation)

    """

    fft_size = _default_fft_size(window_time, fs)
    if f_max is None: f_max = fs/2
    gt_weights, _ = fft_weights(
        fft_size, fs, channels, f_min, f_max=fs/2, maxlen=(fft_size/2 + 1)
        )

    sgram = spectrogram(wave, fft_size, fs, window_time, hop_time)
    result = np.abs(sgram).dot(gt_weights) / fft_size

    return result


def fft_weights(nfft, fs, channels, f_min, f_max, maxlen, width=1):
    """Generate a matrix of weights to combine FFT bins into Gammatone bins.

    Parameters
    ----------
    nfft : int.
        Size of the source FFT.
    fs : int.
        Sampling frequency.
    channels : int.
        Number of output bands required.
    f_min : lower limit of frequencies (Hz)
    f_max : upper limit of frequencies (Hz)
    maxlen : int.
        Number of bins to truncate the columns to.
    width : int; default=1.
        Constant width of each band in Bark.

    Returns
    -------
    (np.ndarray, np.ndarray)
        Weight matrices and gain vectors.
    
    Copyright
    ---------
    (c) 2004-2009 Dan Ellis dpwe@ee.columbia.edu  based on rastamat/audspec.m
    (c) 2012 Jason Heeris (Python implementation)

    """
    ucirc = np.exp(1j*2*np.pi*np.arange(0, nfft/2 + 1)/nfft)[np.newaxis, ...]
    
    # Common ERB filter code factored out
    cf_array = erb_space(f_min, f_max, channels)[::-1]

    _, A11, A12, A13, A14, _, _, _, B2, gain = (
        make_erb_filters(fs, cf_array, width).T
    )
    
    A11, A12, A13, A14 = (A11[..., np.newaxis], A12[..., np.newaxis],
                          A13[..., np.newaxis], A14[..., np.newaxis])

    r = np.sqrt(B2)
    theta = 2 * np.pi * cf_array / fs    
    pole = (r * np.exp(1j * theta))[..., None]
    
    GTord = 4
    
    weights = np.zeros((channels, nfft))

    weights[:, 0:ucirc.shape[1]] = (
          np.abs(ucirc + A11 * fs) * np.abs(ucirc + A12 * fs)
        * np.abs(ucirc + A13 * fs) * np.abs(ucirc + A14 * fs)
        * np.abs(fs * (pole - ucirc) * (pole.conj() - ucirc)) ** (-GTord)
        / gain[..., None]
    )

    weights = weights[:, 0:int(maxlen)]

    return weights.T, gain.T  # Transposed to conform with NEMS shapes.



import itertools

import numpy as np


def correlation(x, y):
    """Compute Pearson correlation coefficient for arrays x and y.

    Correlation is computed for corresponding output channels in `x` and `y`,
    i.e. `corr(x[..., i], y[..., i])` for all `i`.

    TODO: Support for batched data.

    Parameters
    ----------
    x, y : np.ndarray
        Arrays must have the same shape and no more than 2 dimensions.
        Most commonly, these will be a model output (prediction) and a recorded
        response (target).

    Returns
    -------
    float or ndarray.
    
    """
    if (x.ndim > 2) or (y.ndim > 2):
        # TODO: Correlation may not be an appropriate tool for higher-D data?
        #       In which case change this to ValueError, but need to think about
        #       it some more first.
        raise NotImplementedError(
            "`nems.metrics.correlation` only supports 1D or 2D array inputs."
        )

    if x.ndim == 1:
        x = x[...,np.newaxis]
    if y.ndim == 1:
        y = y[...,np.newaxis]

    # NOTE: For large arrays (>100000 time points, >100 neurons) this loop is
    #       much faster than using `np.corrcoef` once and indexing into the full
    #       correlation matrix.
    corrs = []
    for i in range(x.shape[-1]):
        # `rowvar=False` to convert NEMS shapes to shape expected by NumPy.
        # I.e. second index is variables (channels), first index is observations
        # within each variable (time).
        corrs.append(np.corrcoef(x[...,i], y[...,i], rowvar=False)[0,1])

    if len(corrs) == 1:
        # 
        r = corrs[0]
    else:
        r = np.array(corrs)
    
    return r


def noise_corrected_r(prediction, single_trial_responses, n_pairs=None,
                      trial_axis=0, channel_axis=-1):
    """Compute noise-corrected correlation coefficient for a model prediction.

    Noise is determined by measuring single-trial correlations in the recorded
    response. Based on method in Hsu and Theusnissen (2004), Network.

    Parameters
    ----------
    prediction : np.ndarray.
        Typically the output of `Model.predict`.
    single_trial_responses : np.ndarray.
        Recorded neural response, as a single array with shape
        (Trials, Time, Neurons).
    n_pairs : int or None; optional.
        Number of random single trial pairs to test. If None, test all pairs.
    trial_axis : int; default=0.
        Axis to concatenate for arrays in `single_trial_responses`.
        For example, if `trial_axis=0` and
        `single_trial_responses = [array(shape=(1, 50)), array(shape=(1, 50))]`
        then `all_trials = array(2, 50)`.
    channel_axis : int; default=-1.
        Axis for multiple responses. For example,
        if `single_trial_responses = array(shape=(10, 1000, 50))` representing
        responses of 50 neurons to 10 trials of 1000 time units each, then use
        `channel_axis = 2` to compute this metric for each of the 50 neurons.

    Returns
    -------
    corrected_rs : ndarray.
        One value per channel.
        
    """

    n_channels = single_trial_responses.shape[channel_axis]
    
    corrected_rs = []
    for i in range(n_channels):
        one_channel_pred = np.take(prediction, indices=[i],
                                   axis=channel_axis).squeeze()
        one_channel_resp = np.take(single_trial_responses, indices=[i],
                                   axis=channel_axis)
        trial_pair_r = _paired_single_trial_r(one_channel_resp, n_pairs=n_pairs,
                                              trial_axis=trial_axis)
        trial_pred_r = _single_trial_r(one_channel_pred, one_channel_resp,
                                       trial_axis=trial_axis)
        corrected_r= np.mean(trial_pred_r)/np.sqrt(trial_pair_r)
        corrected_rs.append(corrected_r)
    
    return np.array(corrected_rs)
    
def _paired_single_trial_r(trials, n_pairs=1000, trial_axis=0, limit=0.01):
    """Compute mean correlation of pairs of single trials.

    Internal for `noise_corrected_r`.

    Parameters
    ----------
    trials : ndarray
        Single trial responses concatenated along `trial_axis`.
    n_pairs : int or None; optional.
    trial_axis : int; default=0.
    limit : float
        Minimum value to return, to prevent `noise_corrected_r` from
        blowing up.

    Returns
    -------
    mean_pairwise_correlation : float

    """

    if trials.shape[0] == 1:
        # Only 1 repetition
        raise ValueError("`single_trial_responses` must contain more than 1 "
                         "trial to compute `noise_corrected_r`.")

    n_repetitions = trials.shape[trial_axis]
    # Get all pairs of indices into trial axis,
    # then randomly shuffle the pairs
    pairs = list(itertools.combinations(range(n_repetitions), 2))[:n_pairs]
    np.random.shuffle(pairs)

    # Compute correlation of each pair of responses,
    pairwise_correlations = []
    for i, j in pairs:
        rep1 = np.take(trials, indices=[i], axis=trial_axis).squeeze()
        rep2 = np.take(trials, indices=[j], axis=trial_axis).squeeze()
        pairwise_correlations.append(np.corrcof(rep1, rep2)[0, 1])
                
    # Hard limit on single-trial correlation to prevent explosion
    # TODO: better logic for this
    mean_pairwise_correlation = np.mean(pairwise_correlations)
    if mean_pairwise_correlation < limit:
        mean_pairwise_correlation = limit

    return mean_pairwise_correlation

def _single_trial_r(prediction, trials, trial_axis=0):
    """Compute mean correlation between prediction and single trial responses.
    
    Internal for `noise_corrected_r`.

    Parameters
    ----------
    prediction : ndarray
    trials : ndarray
        Single trial responses concatenated along `trial_axis`.
    trial_axis : int; default=0.

    Returns
    -------
    mean_trial_correlation : float
    
    """

    n_repetitions = trials.shape[trial_axis]
    trial_correlations = []
    for i in range(n_repetitions):
        single_trial = np.take(trials, indices=[i], axis=trial_axis).squeeze()
        trial_correlations.append(np.corrcoef(prediction, single_trial)[0, 1])

    return np.mean(trial_correlations)



import numpy as np


class Distribution:
    """Base class for a Distribution."""

    # Any subclass of Distribution will be registered here, for use by
    # `Distribution.from_json`
    subclasses = {}
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses[cls.__name__] = cls

    @classmethod
    def value_to_string(cls, value):
        if value.ndim == 0:
            return str(value)
        else:
            shape = ','.join(str(v) for v in value.shape)
            return f'array(shape=({shape}))'

    def mean(self):
        """Return the expected value of the distribution."""
        return self.distribution.mean()

    def percentile(self, percentile):
        """Calculate the percentile.

        Parameters
        ----------
        percentile : float [0, 1]
            Probability at which the result is calculated. Should be specified as
            a fraction in the range 0 ... 1 rather than a percent.

        Returns
        -------
        value : float
            Value of random variable at given percentile

        Examples
        --------
        For some distributions (e.g., Normal), the bounds will be +/- infinity.
        In those situations, you can request that you get the bounds for the 99%
        interval to get a slightly more reasonable constraint that can be passed
        to the fitter.
        >>> from nems.distributions.api import Normal
        >>> prior = Normal(mu=0, sd=1)
        >>> lower = prior.percentile(0.005)
        >>> upper = prior.percentile(0.995)

        """
        return self.distribution.ppf(percentile)

    @property
    def shape(self):
        return self.mean().shape

    def sample(self, n=1, bounds=None, max_tries=None):
        """Draw random sample(s) from a (truncated) distribution.
        
        Parameters
        ----------
        n : int
            Number of random samples to get.
        bounds : 2-tuple or None
            If not None, samples with at least one value less than `bounds[0]`
            or greater than `bounds[1]` will be rejected and replaced with
            a new sample.

        Returns
        -------
        good_sample : ndarray
        
        """
        if max_tries is None:
            max_tries = 100*n

        size = [n] + list(self.shape)
        good_sample = np.full(shape=size, fill_value=np.nan)

        i = 0
        while np.sum(np.isnan(good_sample)) > 0:
            sample = self.distribution.rvs(size=size)
            if bounds is not None:
                if i > max_tries:
                    raise ValueError(
                        "Resampling within bounds exceeded `max_tries` for "
                        f"{self.tolist()}. with bounds: {bounds}."
                    )
                lower, upper = bounds
                keep = (sample >= lower) & (sample <= upper)
                good_sample[keep] = sample[keep]
            else:
                good_sample = sample
                break
                
        # Drop first dimension if n = 1
        if n == 1:
            good_sample = np.squeeze(good_sample, axis=0)

        return good_sample

    def tolist(self):
        """Represent distribution as a list.
        
        See also
        --------
        Distribution.to_json
        Distribution.from_json

        """
        name = type(self).__name__
        d = self.__dict__.copy()
        for k in list(d.keys()):
            # Remove `self.distribution`
            # and any attributes with two leading underscores
            if (k == 'distribution') or (k.startswith(f'_{name}__')):
                _ = d.pop(k)

        l = [name, d]
        return l

    def to_json(self):
        """Encode a distribution instance as a dictionary.

        See also
        --------
        `nems.tools.json`.
        
        """
        return {'data': self.tolist()}

    @classmethod
    def from_json(cls, json):
        """Decode a distribution from a dictionary.
        
        Warnings
        --------
        Distribution subclasses should avoid assigning instance attributes that
        are not used in `__init__`, as this will break compatibility with
        `from_json`. If such attributes are absolutely needed, use two leading
        underscores (i.e. `self.__my_attr = attr`), as those will not be encoded
        by `to_json`.

        See also
        --------
        `nems.tools.json`.

        """
        class_name, kwargs = json['data']
        # remove first leading underscore kwargs keys, if any
        kwargs = {'_'.join(k.split('_')[1:]): v for k, v in kwargs.items()}
        if class_name == 'Distribution':
            class_obj = Distribution
        else:
            class_obj = cls.subclasses[class_name]
        return class_obj(**kwargs)

    def __eq__(self, other):
        if isinstance(other, Distribution):
            list1 = self.tolist()
            list2 = other.tolist()
            conditions = [
                (list1[0] == list2[0]),  # same subclass
                all([(k1 == k2 and np.allclose(v1, v2)) for (k1,v1), (k2,v2) in
                     zip(list1[1].items(), list2[1].items())])
            ]
            return all(conditions)
        else:
            return NotImplemented



"""Collection of performance metrics and related utilities.

Score model outputs by prediction correlation and other metrics.
Measure equivalence, sparsity, and other properties of model predictions and/or
recorded neural responses.

# TODO: These contents don't all exist yet, sketching this out based on
#       the strfy proposal document.
Contents
--------
    `correlation.py` : As above, but compute Pearson correlation coefficient.
    `equivalence.py` : Measure functional similarity of different models.
    `sparseness.py`  : Measure lifetime sparseness of responses and predictions.

Notes
-----
Asymmetric metrics that compare model outputs to targets
(i.e. `metric(x,y) != metric(y,x)`) should define arguments in the order of:
`metric(output, target)` since that is the order used by `model.score`.

"""

from .correlation import correlation, noise_corrected_r

from nems.tools.lookup import FindCallable

metric_nicknames = {}
get_metric = FindCallable({**globals(), **metric_nicknames}, header='Metric')



from scipy import stats
import numpy as np

from .base import Distribution


class Exponential(Distribution):
    """Exponential Prior.

    Parameters
    ----------
    beta : scalar or ndarray
        Scale of distribution.

    """

    def __init__(self, beta):
        self._beta = np.asarray(beta)
        self.distribution = stats.expon(scale=beta)

    def __repr__(self):
        return 'Exponential(β={})'.format(self._beta)



"""Tools for optimizing NEMS Models with TensorFlow."""

import tensorflow as tf


gpus = tf.config.list_physical_devices('GPU')
if gpus:
  try:
    # Currently, memory growth needs to be the same across GPUs
    for gpu in gpus:
      tf.config.experimental.set_memory_growth(gpu, True)
    logical_gpus = tf.config.list_logical_devices('GPU')
    print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
  except RuntimeError as e:
    # Memory growth must be set before GPUs have been initialized
    print(e)


from .backend import TensorFlowBackend
from .layer_tools import NemsKerasLayer, Bounds



import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras.constraints import Constraint


# TODO: handle frozen parameters
class NemsKerasLayer(Layer):

    def __init__(self, nems_layer, new_values=None, new_bounds=None,
                 regularizer=None, *args, **kwargs):
        """TODO: docs."""
        # Don't pass 'input_shape' to Keras Layer constructor. Will still be in here
        # if `nems_layer.as_tensorflow_layer` didn't use it.
        _ = kwargs.pop('input_shape', None)
        super().__init__(name=nems_layer.name, dtype=nems_layer.parameters.dtype,
                         *args, **kwargs)
        if new_values is None: new_values = {}
        if new_bounds is None: new_bounds = {}
        for p in nems_layer.parameters:
            self._weight_from_nems(
                p, new_values.get(p.name, None), new_bounds.get(p.name, None),
                regularizer
                )

    def _weight_from_nems(self, p, value=None, bounds=None, regularizer=None):
        """Convert Parameter of NEMS Layer to a Keras Layer weight."""
        if value is None: value = p.values
        if bounds is None: bounds = p.bounds
        init = tf.constant_initializer(value)
        constraint = Bounds(bounds[0], bounds[1])
        trainable = not p.is_frozen
        setattr(
            self, p.name, self.add_weight(
                shape=value.shape, initializer=init, trainable=trainable,
                regularizer=regularizer, name=p.name, constraint=constraint
                )
            )

    @property
    def parameter_values(self):
        """Returns key value pairs of the weight names and their values."""
        # TF appendss :<int> to the weight names.
        values = {weight.name.split(':')[0]: weight.numpy()
                    for weight in self.weights} 
        return values

    def weights_to_values(self):
        return self.parameter_values


class Bounds(Constraint):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper
    def __call__(self, w):
        # TODO: This is not a great way to do constraints, but I'm not sure
        #       where alternatives would be implemented (like modifying the
        #       cost function to grow very large when nearing the bound).
        # TODO: Maybe try this version:
        # https://www.tensorflow.org/probability/api_docs/python/tfp/math/clip_by_value_preserve_gradient
        return tf.clip_by_value(w, self.lower, self.upper)



"""Utilities for visualizing the model fitting process."""

import numpy as np
import matplotlib.pyplot as plt

from .model import input_heatmap
from .tools import standardize_axes


def prediction_vs_target(input, target, model, ax=None, sampling_rate=None,
                         show_input=False, xlim=None, title=None,
                         xlabel='Time (bins)', ylabel='Firing Rate (Hz)'):
    """Plot actual target, overlay with model prediction.

    Parameters
    ----------
    input : np.ndarray, dict, or DataSet.
        Input to `model`. See `nems.Model.evaluate`.
    target : np.ndarray.
        Target of optimization. See `nems.Model.fit`.
    model : nems.Model.
        Model to generate prediction with.
    ax : Matplotlib.axes.Axes; optional.
        Axes on which to plot.
    sampling_rate : float; optional.
        If given, label time-axis of some plots with real units (seconds
        by default). See `nems.visualization.tools.ax_bins_to_seconds`.
    show_input : bool; default=False.
        If True, add a heatmap of input above response and prediction.
    xlim : tuple; optional.
        Positional arguments for: `ax.set_xlim(*xlim)`.
    title : str; optional.
    xlabel : str; default='Time (bins)'.
    ylabel : str; default='Firing Rate (Hz)'.

    Returns
    -------
    matplotlib.axes.Axes

    Notes
    -----
    Designed for targets with shape (T, 1). Other shapes may work but the plot
    may be difficult to see.
    
    """

    if ax is None: ax = plt.gca()
    if xlim is None: xlim = (None, None)
    ax.plot(target, c='black', alpha=0.3, label='actual')
    ax.plot(model.predict(input), c='black', label='predicted')
    # Have to do this after each plot to get margins to adjust correctly.
    standardize_axes(ax, sampling_rate=None)

    if show_input:
        # Add input heatmap above prediction
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        extent = [xmin, xmax, ymax*(21/20), ymax*(5/4)]
        input_heatmap(input, ax=ax, extent=extent, add_colorbar=False)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax*(5/4))

    ax.legend(frameon=False, bbox_to_anchor=(1.0, 0.65), loc='lower right')
    # Check not None so that existing title is not erased.
    if title is not None:  ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xlim(*xlim)
    standardize_axes(ax, sampling_rate=sampling_rate)

    return ax


def get_model_errors(model_list):
    """Extract final fit error from FitResults of each Model in list."""
    errors = [m.results.final_error for m in model_list[1:]]
    errors.insert(0, model_list[1].results.initial_error)
    return errors


def iteration_vs_error(model_list, ax=None):
    """Plot error (cost) after each fit iteration.
    
    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.
    ax : Matplotlib.axes.Axes.
        Axes on which to plot.

    Returns
    -------
    matplotlib.axes.Axes

    Notes
    -----
    While the Model at index 0 is not actually used by this function, it is
    included because it is used by related plotting functions like
    `parameter_space_pca`.
    
    """

    if ax is None: ax = plt.gca()
    errors = get_model_errors(model_list)
    ax.plot(errors, c='black')
    ax.set_ylabel('Error')
    standardize_axes(ax)

    return ax


def evals_per_iteration(model_list, ax=None):
    """Plot number of cost function evaluations in each fit iteration.
    
    Only works for models fit using `backend='scipy'`.

    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.
    ax : Matplotlib.axes.Axes.
        Axes on which to plot.

    Notes
    -----
    While the Model at index 0 is not actually used by this function, it is
    included because it is used by related plotting functions like
    `parameter_space_pca`.
    
    """

    try:
        evals_per_iter = [m.results.misc['scipy_fit_result'].nfev
                          for m in model_list[1:]]
        evals_per_iter.insert(0, np.nan)
    except KeyError:
        raise TypeError(
            '`evals_per_iteration` models must be fit with `backend="scipy"`'
            )

    if ax is None: ax = plt.gca()
    ax.plot(evals_per_iter, c='red')
    ax.set_ylabel('Evaluations per iteration')
    ax.set_xlabel('Iteration')
    ax.set_xlim(0, len(model_list)-1)
    standardize_axes(ax)

    return ax


def get_parameter_pcs(model_list):
    """Treat models as observations, project parameters onto PC space.
    
    For a list of M models with N parameters each (i.e. length of
    `Model.get_parameter_vector`), projection matrix will have shape (M, N).
    If each Model (in order) represents the result of a successive fit iteration,
    then each column represents the path of optimization projected onto that
    PC.

    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.

    Returns
    -------
    projections : np.ndarray.
    percent_variance : np.ndarray.
    evecs : np.ndarray

    """

    # Get all parameter vectors as an array with models as observations (rows)
    # and individual parameters as features/variables (columns).
    parameters = np.array([m.results.final_parameters if m.results is not None
                           else m.get_parameter_vector() for m in model_list])

    # Normalize within parameter, compute covariance matrix.
    parameters -= parameters.mean(axis=0)  # mean 0
    parameters /= parameters.std(axis=0)   #  std 1
    cov = np.cov(parameters.T)
    
    # Get eigenvalues and eigenvectors, project parameters and compute 
    # percent variance explained.
    evals, evecs = np.linalg.eigh(cov)  # cov always real, symmetric, P.S.D.
    # Switch from increasing to decreasing order.
    evals = np.flip(evals, axis=0)
    evecs = np.flip(evecs, axis=1)
    # Project parameters on to principal components, get pct variance explained
    projections = parameters @ evecs
    percent_variance = [e/np.sum(evals)*100 for e in evals]

    return projections, percent_variance, evecs


def parameter_space_pca(model_list, ax=None):
    """Plot parameters of each model projected onto first two PCs.

    Principal components are determined by treating each entry in a length-N
    vector returned by `Model.get_parameter_vector()` as a feature, and the
    vector associated with each of the M models in `model_list` as one
    observation, then computing principal components for the resulting
    M x N matrix.
    
    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.
    ax : Matplotlib.axes.Axes.
        Axes on which to plot.

    Returns
    -------
    matplotlib.axes.Axes

    """

    pcs, percent_variance, _ = get_parameter_pcs(model_list)
    errors = get_model_errors(model_list)

    if ax is None: ax = plt.gca()
    ax.plot(pcs[:,0], pcs[:,1], c='black', zorder=-1)
    ax.scatter(pcs[:,0], pcs[:,1], c=errors, s=100)
    ax.scatter(pcs[0,0], pcs[0,1], marker='o', edgecolors='red',
               facecolors='none', s=200, label='Initial')
    ax.scatter(pcs[-1,0], pcs[-1,1], marker='D', edgecolors='red',
               facecolors='none', s=200, label='Final')

    colors = ax.get_children()[1]
    plt.colorbar(colors, label='Error', ax=ax)
    ax.legend(frameon=False)
    ax.set_xlabel(f'Parameter PC1 ({percent_variance[0]:.1f}% var)')
    ax.set_ylabel(f'Parameter PC2 ({percent_variance[1]:.1f}% var)')
    ax.set_xticks([])
    ax.set_yticks([])
    standardize_axes(ax, x_margin=True)

    return ax


def parameter_pca_table(model_list, n=None, m=3, ax=None, fontsize=16):
    """Generate a table showing Parameters with top 5 weights from first n PCs.
    
    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.
    n : int; optional.
        Number of PCs to show. All shown by default.
    m : int; default=3.
        List Parameters with the top `m` weights.
    ax : Matplotlib.axes.Axes.
        Axes on which to plot.

    Returns
    -------
    matplotlib.axes.Axes

    """

    _, percent_variance, evecs = get_parameter_pcs(model_list)
    first_n_pcs = evecs[:,:n].T
    top_m_idx = [np.flip(np.argsort(np.abs(e)))[:m] for e in first_n_pcs]
    row_labels = [f'PC{i} ({percent_variance[i-1]:.1f}%)' for i in range(1,n+1)]
    table_text = []

    for i, (ev, idx) in enumerate(zip(first_n_pcs, top_m_idx)):
        table_text.append(['']*m)
        parameters = [
            f'{k}: {p.name}' for k, p in
            model_list[0].get_parameter_from_index(*idx).items()
            ]

        for j, p in enumerate(parameters):
            table_text[i][j] = (p + f';   {ev[j]:.3f}')

    if ax is None: ax = plt.gca()
    ax.imshow([[0]], cmap='Greys')
    table = ax.table(table_text, rowLabels=row_labels, loc='center')
    table.auto_set_font_size(False)
    table.set_fontsize(fontsize)
    table.scale(fontsize/10, fontsize/10)
    ax.axis('off')
    ax.axis('tight')

    return ax


def scipy_gradient(model_list, ax=None):
    """Plot L2-norm of gradient for each model.
    
    Parameters
    ----------
    model_list : list of nems.Model.
        Must contain at least 2 Models. The Model at index 0 is interpreted as
        the "original" (un-fit) Model. Other models are interpreted as the
        results of a successive fit iterations. In particular, `model.results`
        must not be None for Models at indices 1 or greater.
    ax : Matplotlib.axes.Axes.
        Axes on which to plot.

    Returns
    -------
    matplotlib.axes.Axes

    """

    grads = np.array([m.results.misc['scipy_fit_result'].jac
                      for m in model_list[1:]])
    # Use all nan gradient for first iteration (not fit).
    all_nan = np.full(shape=(1, grads.shape[1]), fill_value=np.nan)
    grads = np.concatenate([all_nan, grads])
    norm = np.sqrt(np.sum(grads**2, axis=1))

    if ax is None: ax = plt.gca()
    ax.plot(norm, c='black')
    ax.set_xlabel('Iteration')
    ax.set_ylabel('L2 Norm of Gradient')
    ax.set_xlim(0, len(model_list)-1)
    standardize_axes(ax)

    return ax



import copy
import textwrap
import itertools
import warnings

import numpy as np

from nems.registry import keyword_lib
from nems.backends import get_backend
from nems.metrics import get_metric
from nems.visualization import plot_model
from nems.tools.arrays import one_or_more_nan
from .dataset import DataSet
# Temporarily import layers to make sure they're registered in keyword_lib
import nems.layers
del nems.layers


class Model:

    def __init__(self, layers=None, name=None, dtype=np.float64, meta=None):
        """A structured collection of Layers.
        
        This is the primary class for interacting with NEMS. Conceptually, a
        Model encapsulates all computations needed to transform an input into
        a desired output (or prediction).

        TODO: more context here?

        Parameters
        ----------
        layers : list of Layer; optional.
            Layers that will define the Model's data transformations.
        name : str; optional.
            Name for the Model.
        dtype : type; default=np.float64.
            TODO: docs. float64 not supported by all TensorFlow ops.
        meta : dict; optional.
            A general-purpose dictionary for storing additional information
            about the model.

        Attributes
        ----------
        layers : _LayerDict.
            All model layers in a format that allows both integer and
            string indexing.
        results : FitResults or None.
            A collection of information about the most recent call of
            `Model.fit()`. This will be `None` if `Model.fit()` has never been
            called for this Model.
        backend : Backend or None.
            A container for an equivalent model built using TensorFlow or some
            other supported backend, cached by the most recent call to
            `Model.fit()`. This will be `None` if `Model.fit()` has never been
            called.

        Methods
        -------
        evaluate(input, ...)
            Transform input by applying all Layers in a proscribed order.
        predict(input, ...)
            As `evaluate`, but will only return the output of the final layer
            by default.
            NOTE: Currently this method assumes there is only one model output
                and a single target.
        fit(input, target, ...)
            Produce a copy of this Model with parameters optimized to match
            the output of `Model.evaluate(input)` to `target`.
        plot(input, target=None, ...)
            Visualize the output of each Layer when applied to `input` (or the
            output of a previous Layer). If `target` is also specified, plot
            `target` on the same axes as the output of the final Layer.
            NOTE: Currently this method assumes there is only one model output
                  and a single target.

        See also
        --------
        nems.layers.base.layer.Layer
        nems.backends.base.Backend
        nems.visualization.model.plot_model

        Examples
        --------
        >>> import numpy as np
        >>> from nems import Model
        >>> # Create some fake data
        >>> input = np.random.rand(1000,18)  # 1000 time points, 18 channels
        >>> target = np.random.rand(1000, 1)
        >>> # Compute the model's output given the input data.
        >>> evaluated_data = model.evaluate(input)
        >>> # Get a fitted model
        >>> fit_model = model.fit(input, target)
        >>> # Plot the fitted model
        >>> fig = fit_model.plot(input, target)
        
        """
        self._layers = {}  #  layer.name : layer obj, increment on clashes

        # TODO: remove warning after fixing issues w/ scipy
        if dtype != np.float64:
            warnings.warn(
                "Using `Model(dtype=...)` is currently experimental. For best "
                "results, leave dtype as the default. TF Backend will overwrite "
                "this setting to np.float32 for the time being."
            )
        self.dtype = dtype

        if layers is not None:
            self.add_layers(*layers)
        self.name = name if name is not None else 'UnnamedModel'

        # Store optional metadata. This is a generic dictionary for information
        # about the model. Any type can be stored here as long as it can be
        # encoded by `json.dumps`.
        if meta is None: meta = {}
        self.meta = meta

        self.results = None   # holds FitResults after Model.fit()
        self.backend = None # holds all previous Backends (1 per key)

    @property
    def layers(self):
        """Get all Model Layers. Supports integer or string indexing."""
        return _LayerDict(self._layers)

    @property
    def bounds(self):
        """Get all Model bounds as a dict. See `Layer.bounds`."""
        return {k: v.bounds for k, v in self.layers.items()}

    @property
    def priors(self):
        """Get all Model priors as a dict. See `Layer.priors`."""
        return {k: v.priors for k, v in self.layers.items()}

    @property
    def parameter_count(self):
        """Total size of all Parameters from all Layers.
        
        Note that this is the total number of all parameter values, *not* the
        number of Parameter objects. I.e. a model with a single Parameter of
        shape (2,3) has a parameter_count of 6.
        TODO: rename this to avoid ambiguity? value_count? 

        Returns
        -------
        int

        See also
        --------
        Model.parameter_info
        nems.layers.base.Layer.parameter_count
        
        """
        return sum([layer.parameter_count for layer in self.layers])

    @property
    def parameter_info(self):
        """Sizes of frozen, unfrozen and permanent parameters.
        
        Returns
        -------
        dict
            {'layer_name':  # per layer
                {'total': int, 'unfrozen': int, 'frozen': int, 'permanent': int},
                ...
             'model':       # model totals
                {'total': int, unfrozen': int, ... }  # etc.
                }

        See also
        --------
        Model.parameter_count
        nems.layers.base.Layer.parameter_info
        
        """
        info = {layer.name: layer.parameter_info for layer in self.layers}
        model_info = {k: sum([d[k] for d in info.values()])
                      for k in list(info.values())[0]}
        info['model'] = model_info

        return info

    def set_dtype(self, dtype):
        """Change dtype of Parameters in-place, set dtype for fit/evaluate."""
        for layer in self.layers:
            layer.set_dtype(dtype)
        self.dtype = dtype

    def add_layers(self, *layers):
        """Add Layers to this Model, stored in `Model._layers`.

        This will also update `Layer.name` for any layers with a name clash,
        so that each Layer in the Model is guaranteed to have a unique name.

        Parameters
        ----------
        layers : N-tuple of Layers

        See also
        --------
        nems.layers.base.Layer
        
        """

        # TODO: need to track name, layer lists instead? Apparently dictionaries
        #       aren't guaranteed to keep the same order. Hasn't caused problems
        #       so far, but...

        for layer in layers:
            layer.model = self  # each layer gets a reference to parent Model
            key = layer.name
            i = 0
            while key in self._layers:
                # Avoid name clashes
                key = f'{layer.name}{i}'
                i += 1
            self._layers[key] = layer
            # Also update `Layer.name` so that there's no mismatch between
            # a Layer's name and its key in the Model.
            layer._name = key

        self.set_dtype(self.dtype)

    def get_layer_index(self, name):
        """Get integer index for Layer with `.name == name`."""
        return list(self.layers.keys()).index(name)

    def get_layer(self, key):
        """Get one Layer. Key can be int or string (`Layer.name`)."""
        return self.layers[key]

    # Maybe we don't need to implement this? Would require some refactoring of
    # Model.layers.
    def insert_layer(self, index, name=None):
        """TODO: add layer at specific integer index."""
        raise NotImplementedError

    def evaluate(self, input, state=None, input_name=None, state_name=None,
                 output_name=None, n=None, return_full_data=True,
                 as_dataset=False, use_existing_maps=False, batch_size=0,
                 permute_batches=False, debug_nans=False):
        """Transform input(s) by invoking `Layer.evaluate` for each Layer.

        Evaluation encapsulates three steps:
            1) Package data and metadata in a single structured container.
            2) Loop over `Model.layers`, invoking `Layer._evaluate` to
               transform the data.
            3) Clean up no-longer-needed data, possibly re-format.
        See `DataSet` and `Model.generate_layer_data` for implementation.

        During the evaluation process, `input` (and `state` if provided) will
        be packaged into dictionaries, with the following structure:
            array_name : ndarray. 
                If `input` is a dict, each array in `input` will be
                shallow-copied with the same key. Otherwise, arrays `input` and
                `state` will be added to with default keys specified by
                the DataSet class.
            _last_output : ndarray, list, or None
                Return value of the most recently evaluated Layer. This key is
                removed after evaluation is complete.
        These dictionaries are tracked separately within the DataSet object
        for inputs, outputs, and targets.

        Parameters
        ----------
        input : ndarray, dict, or DataSet.
            If ndarray, use this as the input to the first Layer.
            Otherwise, use keys specified by `input_name` or `Layer.input` to
            determine the first input.
        state : ndarray; optional.
            Add this to `DataSet.inputs`. This option can only be used in
            conjunction with an array `input`. If other data is needed,
            use a dictionary input containing all data.
        input_name : str; optional.
            Specifies which array should be provided as input to the
            first layer. Note that priority will be given to `Layer.input` in
            case of a clash. I.e. if `input_name is not None`, but also
            `Layer.input is not None` for the first layer, `Layer.input` will
            be used.
            If None : use `Layer.input` of first layer if specified.
            If str  : a single input array at key `input_name`.
        state_name : str; optional.
            If str, and `state is not None`, then `state_name` will be the key
            for `state`.
        output_name : str; optional.
            Specifies name(s) for output(s) of the final layer if
            `Layer.output is None` (incremented if multiple).
        n : int; optional.
            Evaluate the first `n` Layers (all by defualt).
        return_full_data : bool; default=True
            If True, return a dictionary containing all input data and all
            uniquely-keyed Layer outputs.
        as_dataset : bool; default=False.
            Return data in a DataSet instead of a dictionary.
        use_existing_maps : bool; default=False.
            If True, existing DataMaps will be used rather than generating a
            new one for each Layer. This is disabled by default for direct
            evaluation, but enabled by default when called from `Model.fit` to
            eliminate unnecessary overhead for repeated evaluations of the same
            data.
        batch_size : int; default=0.
            Determines batching behavior.
            If 0 (default) : Data will be interpreted as having no sample
                dimension (i.e. first dimension is time). This is equivalent to
                using batch_size=1 with a single sample.
            If None : First dimension is interpreted as samples instead of time.
                Use a single batch with batch_size equal to the size of the
                sample dimension.
            Else : Form batches by select groups of samples such that each batch
                includes `batch_size` samples. If the number of samples is not
                evenly divisible by `batch_size`, then one batch will have
                fewer samples.
            Each sample within each batch will be evaluated separately, then
            their outputs will be concatenated to match the shape of the input.
        permute_batches : bool; default=False.
            TODO, still WIP. Currently doesn't track how indices are shuffled
            so the outputs won't be concatenated in the right order.
            If True, randomly shuffle batches prior to evaluation. Typically
            used during fitting, to shuffle between epochs.
        debug_nans : bool; default=False.
            If True, raise ValueError if a Layer's output contains NaN values.
            Useful for narrowing down the source of unexpected NaNs in the final
            model output.

        Returns
        -------
        data : ndarray, list, dict or DataSet
            Type depends on `return_full_data` and `as_dataset` options, and the
            return type of the `evaluate` method of the final layer in the model.

        See also
        --------
        nems.layers.base.Layer._evaluate
        nems.models.dataset.DataSet
        Model.generate_layer_data

        Warnings
        --------
        Since arrays in `data` share memory with `input`, modifying arrays
        arrays in-place is strongly discouraged.
        
        """

        if not isinstance(input, DataSet):
            # Package arrays and/or dicts in a DataSet
            data = DataSet(
                input=input, state=state, input_name=input_name,
                state_name=state_name, output_name=output_name,
                dtype=self.dtype
            )
        else:
            # Input should already be a properly formatted DataSet
            # (for example, passed by Model.fit)
            data = input

        if batch_size == 0:
            # No reason to do any batching, data should be formatted as
            # (T, ..., N), where T is the number of time bins and N is the
            # number of output channels.
            data_generator = self.generate_layer_data(
                                data, use_existing_maps=use_existing_maps,
                                debug_nans=debug_nans
                                )
            data = self.get_layer_data(data_generator, n, n)[-1]['data']

        else:
            # Data should be formatted as (S, T, ..., N), where S is the number
            # of samples/trials.

            batch_out = list(self.generate_batch_data(
                input, state=state, input_name=input_name, state_name=state_name,
                output_name=output_name, n=n, batch_size=batch_size,
                permute_batches=permute_batches, debug_nans=debug_nans,
                use_existing_maps=use_existing_maps
                ))
            all_outputs = DataSet.concatenate_sample_outputs(batch_out)
            # Inputs (and any targets) should not have changed
            data.outputs = all_outputs

        if not return_full_data:
            out = data.outputs
            if len(out) == 1:
                out = list(out.values())[0]
        else:
            if as_dataset:
                out = data
            else:
                # Default option: return all data in a dictionary
                out = data.as_dict()

        return out

    def generate_batch_data(self, input, n=None, batch_size=0,
                            permute_batches=False, use_existing_maps=False,
                            **eval_kwargs):
        """Generate output of final Layer for one batch at a time.
        
        See `Model.evaluate` for detailed documentation of parameters.

        Parameters
        ----------
        input : ndarray, dict, or DataSet
        n : int; optional.
        batch_size : int or None; default=0.
        permute_batches : bool; default=False.
        use_existing_maps : bool; default=False.
        eval_kwargs : dict.
            Additional keyword arguments for `Model.evaluate`.

        Yields
        ------
        batch : DataSet
            Contains all inputs and outputs (and targets if present) for a
            single batch.
        
        """
    
        if not isinstance(input, DataSet):
            # Package arrays and/or dicts in a DataSet
            data = DataSet(input=input, **eval_kwargs)
        else:
            # Input should already be a properly formatted DataSet
            # (for example, passed by Model.fit)
            data = input

        if n is not None: n -= 1
        else: n = len(self.layers)-1

        batches = data.as_batches(batch_size, permute=permute_batches)
        for batch in batches:
            samples = batch.as_samples()
            sample_out = []
            for sample in samples:
                data_generator = self.generate_layer_data(
                    sample, use_existing_maps=use_existing_maps,
                    **eval_kwargs
                )

                # Get data for the final layer only, to reduce memory use.
                layer_data = self.get_layer_data(data_generator, n, n)[-1]
                sample_out.append(layer_data['data'].prepend_samples())
                # Make sure memory is freed up between loops.
                del data_generator
                del layer_data

            batch_out = DataSet.concatenate_sample_outputs(sample_out)
            batch.outputs = batch_out
            yield batch


    # TODO: possibly move this method and any related subroutines to a separate
    #       module (inside a new `base` directory), with simple wrappers in
    #       Model as the public-facing API.
    def generate_layer_data(self, input, copy_data=False, debug_nans=False,
                            use_existing_maps=False, **eval_kwargs):
        """Generate input and output arrays for each Layer in Model.
        
        This method serves as the core loop for `Model.evaluate`, but is exposed
        separately for use in plotting, debugging, etc. The loop is implemented
        as a generator to reduce memory overhead when only one Layer at a time
        is needed.

        Parameters
        ----------
        input : dict or ndarray
            See `Model.evaluate`.
        copy_data : bool; default=False.
            If True, a deep copy of data will be stored in `layer_data['data']`
            after each `Layer._evaluate`. This will be very memory intensive
            for large data, and is generally not recommended except for
            debugging.
        use_existing_maps : bool; default=False.
            If True, existing DataMaps will be used rather than generating a
            new one for each Layer. This is disabled by default for direct
            evaluation, but enabled by default when called from `Model.fit` to
            eliminate unnecessary overhead for repeated evaluations of the same
            data.
        eval_kwargs : dict
            Additional keyword arguments for `Model._initialize_data` and
            `Model._finalize_data`. See `Model.evaluate` for documentation.

        Yields
        ------
        layer_data : dict
            `layer_data` has the following structure: {
                'index': int, index of Layer within Model.
                'layer': str, Layer.name.
                'args': list of ndarray, positional arguments
                    for `Layer._evaluate`.
                'kwargs': dict of ndarray, keyword arguments
                    for `Layer._evaluate`
                'out': ndarray or list of ndarray, return value of
                    `Layer._evaluate(*args, **kwargs)`.
                'data' : DataSet.
                    See `Model.evaluate` for details.
                }

        Warnings
        --------
        layer_data['data'], is a reference to a data structure that is
        iteratively updated in-place during evaluation. Modifying this
        structure in-place is strongly discouraged, as it can violate the state
        expectations of not-yet-evaluated Layers. To make modifications safely,
        use `copy_data=True`.
        TODO: I think this is partially untrue now since each DataSet should
              be a shallow copy of the previous one. Modifying the underlying
              arrays is still discouraged, however. Need to test the first
              point.

        See also
        --------
        Model.get_layer_data
        Model.print_layer_data

        Examples
        --------
        Get a list of all outputs in memory simultaneously:
        >>> generator = generate_layer_data(input, **eval_kwargs)
        >>> data_list = [d['out'] for d, _ in generator]

        Get positional arguments for the first layer:
        >>> generator = generate_layer_data(input, **eval_kwargs)
        >>> args = next(generator)['args']

        """

        if not isinstance(input, DataSet):
            data = DataSet(input, **eval_kwargs)
        else:
            data = input

        max_n = len(self.layers)
        previous_output = 'not None'  # Only for setting `inplace_ok` flag.
        for n, layer in enumerate(self.layers):
            if not use_existing_maps:
                layer.reset_map()

            inplace_ok = False
            # If previous output was None, then it will be overwritten after
            # this Layer is evaluated anyway. But also need to make sure input
            # arrays are not overwritten.
            if previous_output is None:       # never True for first layer.
                if layer.input is None:       # so this must point to an output
                    inplace_ok = True
                else:
                    # Need to check if any of the keys in `args` or `kwargs`
                    # points to an input rather than an intermediate output.
                    arg_keys = layer.data_map.args
                    kwarg_keys = list(layer.data_map.kwargs.values())
                    keys = [*arg_keys, *kwarg_keys]
                    if all([k not in data.inputs for k in keys]):
                        inplace_ok = True

            a, k, o = self._evaluate_layer(layer, data, inplace_ok=inplace_ok,
                                           debug_nans=debug_nans)

            # TODO: last thing to consider (I think) on `inplace_ok`: overwriting
            #       previous output would change the reference to `o` in the
            #       already-yielded layer data. This wouldn't hurt anything in
            #       the model.evaluate implementation b/c only the final data
            #       is used anyway. However, that could make debugging confusing.
            layer_data = {
                'index': n, 'layer': layer.name,
                'args': a, 'kwargs': k, 'out': o
                }

            if n < (max_n - 1):
                if copy_data:
                    layer_data['data'] = copy.deepcopy(data)
                else:
                    layer_data['data'] = data
                yield layer_data
            
            # Track last output for determining `inplace_ok`.
            previous_output = layer.output

        # On final layer, only update data after evaluation
        data.finalize_data(final_layer=layer)
        if copy_data:
            layer_data['data'] = copy.deepcopy(data)
        else:
            layer_data['data'] = data

        yield layer_data

    def _evaluate_layer(self, layer, data, inplace_ok=False, debug_nans=False):
        """Evaluates one Layer. Internal for `Model.generate_layer_data`.
        
        Returns
        -------
        args : list of ndarray
            Positional arguments for `Layer.evaluate`.
        kwargs : dict of ndarray
            Keyword arguments for `Layer.evaluate`.
        output : ndarray or list of ndarray
            Return value of `Layer.evaluate(*args, **kwargs)`.
        
        """
        
        # Get input & output arrays
        args, kwargs, output = layer._evaluate(
            data, inplace_ok=inplace_ok, dtype=self.dtype,
            debug_nans=debug_nans
            )

        # Save output (or don't) based on Layer.DataMap.
        # data_keys is always a list, but output might be a list or one array.
        data_keys = layer.data_map.out
        data.save_output(data_keys, output)

        return args, kwargs, output

    # TODO: maybe remove the data_generator arg and just have this wrap
    #       generate_layer_data? 
    def get_layer_data(self, data_generator, first_index=None, last_index=None):
        """Return data for layers between specified indices (inclusive).
        
        Parameters
        ----------
        data_generator : generator
            Return value of `Model.generate_layer_data`.
        first_index : int; optional.
            Index within Model of the first Layer to get data for.
        last_index : int; optional.
            Index within Model of the last Layer to get data for.

        Returns
        -------
        list of (dict, dict)
            See `Model.generate_layer_data`.

        Examples
        --------
        Get a list of all inputs & outputs in memory simultaneously:
        >>> generator = generate_layer_data(input, **eval_kwargs)
        >>> data_list = get_layer_data(generator)

        Get the keyword arguments for the 3rd Layer:
        >>> generator = generate_layer_data(input, **eval_kwargs)
        >>> kwargs3 = get_layer_data(generator, 3, 3)['kwargs']
        
        """
        if last_index is not None: last_index += 1
        subset = itertools.islice(data_generator, first_index, last_index)
        return [d for d in subset]

    def print_layer_data(self, input, max_char=79, max_array_length=20,
                         show_full_data=False, **eval_kwargs):
        """Pretty-print the return value of `Model.generate_layer_data`.

        Useful for debugging.

        Parameters
        ----------
        input : ndarray, dict, or DataSet.
            See `Model.evaluate`.
        max_char : int; default=79.
            Maximum number of characters to display on each line.
            TODO: separators currently ignore this.
        max_array_length : int; default=20.
            Show truncated arrays if they contain more than this many entries.
            Equivalent to `np.set_printoptions(threshold=max_array_length)`,
            but the previous threshold will be reset after printing.
            TODO: add precision option?
        show_full_data : bool; default=False.
            If True, print the entire DataSet after for each Layer.

        TODO: option to return string instead of printing?
        
        """
        def wrap(k, v):
            return textwrap.fill(f'{k}: {str(v)}', max_char) + '\n' + '-'*16

        current_threshold = np.get_printoptions()['threshold']
        np.set_printoptions(threshold=max_array_length)

        for d in self.generate_layer_data(input, **eval_kwargs):
            _data = d.pop('data')

            # Input/output info
            print('_'*36 + f'in/out:' + '_'*36)
            for k, v in d.items():
                if isinstance(v, list):
                    print(f'{k}:')
                    i = 0
                    for val in v:
                        print(wrap(i, val))
                        i += 1
                    if i == 0:
                        print('-'*16)
                elif isinstance(v, dict):
                    print(f'{k}:')
                    j = 0
                    for key, value in v.items():
                        print(wrap(key, value))
                        j += 1
                    if j == 0:
                        print('-'*16)
                else:
                    print(wrap(k, v))
            print('\u203E'*79)

            if show_full_data:
                # Data dictionary
                print('_'*38 + 'data' + '_'*37)
                for k, v in _data.items():
                    print(wrap(k, v))
                print('\u203E'*79 + '\n\n')

        np.set_printoptions(threshold=current_threshold)


    def get_data_maps(self):
        """Get a dictionary of {Layer.name: Layer.DataMap} for all Layers.

        Similar to `Model.layers`, this dictionary is wrapped so that indexing
        with integers, slices, or multiple keys is also possible.
        
        Returns
        -------
        dict

        See also
        --------
        nems.layers.base.map.DataMap
        
        """
        return _LayerDict({layer.name: layer.data_map for layer in self.layers})

    def predict(self, input, return_full_data=False, **eval_kwargs):
        """As `Model.evaluate`, but return only the last output by default.
        
        TODO: This only works for models where the final layer produces the
              only output. Need to think about how to make it work for models
              that produce multiple outputs at different stages.

              Rough idea: return all arrays from data that were not present
              in input.
        
        """
        return self.evaluate(input, return_full_data=return_full_data,
                             **eval_kwargs)

    def fit(self, input, target, target_name=None, backend='scipy',
            fitter_options=None, backend_options=None, **eval_kwargs):
        """Optimize model parameters to match `Model.evaluate(input)` to target.
        
        TODO: where do jackknife indices fit in? possibly use segmentor idea
              from old NEMS that never really got implemented, as an alternative
              to requiring jackknifing to be set up ahead of time.

              simplest solution: don't deal with it here at all. can have a
              separate method/function that just loops through calls to .fit
              and sets indices in between.

        TODO: want to add explicit support for multiple targets,
              E.g. fit to two types of data simultaneously.

        Parameters
        ----------
        input : np.ndarray or dict.
            See `Model.evaluate`.
        target : np.ndarray or dict of np.ndarray.
            TODO: support dict
            If ndarray, this will be the fitter's target data (i.e. try to
            match the model prediction to this). If dict, ... 
            TODO: dict version is more complicated than I was thinking of.
            Would need to also specify a mapping of output -> target.
        target_name : str; optional.
            Key to assign to target, if target is an ndarray.
        backend : str; default='scipy'.
            Determines how the Model will be fit.
            If 'scipy' : Use `scipy.optimize.minimize(method='L-BFGS-B')`.
            If 'tf'    : Use TensorFlow. Also aliased as 'tensorflow'.
            TODO: any other options we want to support?
            See `nems.backends`.
        fitter_options : dict; optional.
            Keyword arguments to pass on to the fitter. For a list of valid
            options, see documentation for `Backend._fit` in the relevant
            Backend subclass.
        backend_options : dict; optional.
            Keyword arguments to pass to the Backend constructor.
        eval_kwargs : dict
            Keyword arguments to supply to `Model.evaluate`.

        Returns
        -------
        new_model : Model
            A copy of this Model with updated parameters. The Backend object
            used during the fit will be saved in `Model.backend`, and a
            FitResults object will be saved in `Model.results`.

        """

        if fitter_options is None: fitter_options = {}
        if backend_options is None: backend_options = {}

        # Initialize DataSet
        data = DataSet(
            input, target=target, target_name=target_name, **eval_kwargs
            )
        if eval_kwargs.get('batch_size', 0) != 0:
            # Broadcast prior to passing to Backend so that those details
            # only have to be tracked once.
            data = data.as_broadcasted_samples()
        # Evaluate once prior to fetching backend, to ensure all DataMaps are
        # up to date and include outputs.
        _ = self.evaluate(
            input, use_existing_maps=False, **eval_kwargs
            )

        # Update parameters of a copy, not the original model.
        new_model = self.copy()
        # Get Backend sublass.
        backend_class = get_backend(name=backend)
        # Build backend model.
        backend_obj = backend_class(
            new_model, data, eval_kwargs=eval_kwargs,
            **backend_options
            )
        # Fit backend, save results.
        fit_results = backend_obj._fit(
            data, eval_kwargs=eval_kwargs, **fitter_options
            )
        new_model.backend = backend_obj
        new_model.results = fit_results

        return new_model


    def score(self, input, target, metric='correlation', metric_kwargs=None,
              **eval_kwargs):
        """Score model performance using post-fit metrics like correlation.
        
        This only supports metrics that expect a model output as a first
        argument, a target as a second argument, and no other positional
        arguments.

        Parameters
        ----------
        input : ndarray, dict, or DataSet.
            See `Model.evaluate`.
        target : ndarray or dict of ndarray.
            Dict option still TODO. Target data for optimization.
        metric : str or callable; default='correlation'.
            Scores model performance. If `str`, will be replaced with
            `get_metric(metric)`.
        metric_kwargs : dict; optional.
            Additional keyword arguments for `metric`.

        Returns
        -------
        numeric.
            Specifically, `metric(prediction, target, **metric_kwargs)`.
            No specific return format is enforced, but in general metrics should
            return a numeric value, typically a float or ndarray.
        
        """

        if metric_kwargs is None: metric_kwargs = {}
        prediction = self.predict(input, **eval_kwargs)
        if isinstance(metric, str):
            metric = get_metric(metric)

        return metric(prediction, target, **metric_kwargs)


    def get_bounds_vector(self, none_for_inf=True):
        """Get all parameter bounds, formatted as a list of 2-tuples.

        Parameters
        ----------
        none_for_inf : bool
            If True, +/- np.inf is replaced with None
            (for scipy-compatible bounds).

        Returns
        -------
        model_bounds : list of 2-tuple

        See also
        --------
        nems.layers.base.Layer.get_bounds_vector

        """
        # collect all bounds, flatten the intermediate bounds lists
        bounds = [b for layer in self.layers for b in
                  layer.get_bounds_vector(none_for_inf=none_for_inf)]
        return bounds

    def get_parameter_vector(self, as_list=True):
        """Get all parameter values, formatted as a single vector.
        
        Parameters
        ----------
        as_list : bool
            If True, returns a list instead of ndarray
            (for scipy compatibility)

        Returns
        -------
        model_vector : list or ndarray

        See also
        --------
        nems.layers.base.Layer.get_parameter_vector

        """
        # collect all layer vectors
        vectors = []
        for layer in self.layers:
            vector = layer.get_parameter_vector(as_list=as_list)
            vectors.append(vector)

        # flatten list
        if as_list:
            model_vector = [v for vector in vectors for v in vector]
        else:
            model_vector = np.concatenate(vectors)
        
        return model_vector

    def get_parameter_from_index(self, *indices):
        """Get reference(s) to Parameter(s) from index in parameter vector.
        
        Parameters
        ----------
        indices : N-tuple of int.
            Indices corresponding to result of `Model.get_parameter_vector()`.
        
        Returns
        -------
        dict of nems.layers.base.Parameter
            With structure {Layer.name : Parameter}
        
        """
        # collect all layer vectors
        layer_counts = {}
        for layer in self.layers:
            layer_counts[layer.parameter_count] = layer

        parameters = {}
        for i in indices:
            j = 0
            for count, layer in layer_counts.items():
                j += count
                if i >= j:
                    # Parameter is not in this layer
                    continue
                else:
                    # Parameter is in this layer
                    parameters[layer.name] = layer.get_parameter_from_index(
                        i-(j-count)
                        )
                    break

        return parameters

    def set_parameter_vector(self, vector, ignore_checks=False):
        """Set all parameter values with a single vector.

        Parameters
        ----------
        vector : ndarray or list
            New parameter vector. Size must match the total number of flattened
            parameter values.
        ignore_checks : bool
            If True, set new values without checking size or bounds.
            (intended as a minor optimization for the scipy fitter).

        See also
        --------
        nems.layers.base.Layer.set_parameter_vector
        
        """

        first_index = 0
        for layer in self.layers:
            parameter_size = layer.parameters.unfrozen_size
            last_index = first_index + parameter_size
            layer.set_parameter_vector(vector[first_index:last_index],
                                       ignore_checks=ignore_checks)
            first_index = last_index

    def get_parameter_values(self, *layer_keys):
        """Get all parameter values, formatted as a dict.
        
        Parameters
        ----------
        layer_keys : N-tuple of strings
            Keys indicating which Layers to get parameter values for. If no keys
            are specified, get values for all layers.

        Returns
        -------
        all_values : dict

        See also
        --------
        nems.layers.base.Layer.get_parameter_values
        
        """
        if layer_keys == ():
            layer_keys = self._layers.keys()
        all_values = {}
        for k in layer_keys:
            values = self.layers[k].get_parameter_values(as_dict=True)
            all_values[k] = values
        return all_values

    def set_parameter_values(self, layer_dict):
        """Set new parameter values from key, value pairs.
        
        See also
        --------
        nems.layers.base.Layer.set_parameter_values

        """
        for k, v in layer_dict.items():
            self.layers[k].set_parameter_values(v)

    def sample_from_priors(self, n=1):
        """Get a copy of `Model` with new parameter values sampled from priors.
        
        Parameters
        ----------
        n : int; default=1.
            For `n > 1`, a list of `n` Model copies will be returned. This
            option is ignored if `inplace=True`.

        Returns
        -------
        Model, list of Model.

        See also
        --------
        nems.layers.base.Layer.sample_from_priors

        """

        m = self
        model_copies = []
        for i in range(n):
            # Can't copy all from self, otherwise samples will be the same.
            m = m.copy()
            for layer in m.layers:
                layer.sample_from_priors(inplace=True)
            model_copies.append(m)
        if n == 1:
            # Remove singleton list wrapper
            model_copies = model_copies[0]

        return model_copies

    def mean_of_priors(self, n=1):
        """Get a copy of `Model` with parameters set to mean of priors..
        
        Parameters
        ----------
        n : int; default=1.
            For `n > 1`, a list of `n` Model copies will be returned. This
            option is ignored if `inplace=True`.

        Returns
        -------
        Model, list of Model.

        See also
        --------
        nems.layers.base.Layer.mean_of_priors

        """

        model_copies = [self.copy() for i in range(n)]
        for m in model_copies:
            for layer in m.layers:
                layer.mean_of_priors(inplace=True)
        if n == 1:
            # Remove singleton list wrapper
            model_copies = model_copies[0]

        return model_copies

    def set_index(self, index, new_index='initial'):
        """Change which set of parameter values is referenced.

        Intended for use with jackknifing or other procedures that fit multiple
        iterations of a single model. Rather than using many copies of a full
        Model object, each layer tracks copies of its parameter values.

        Parameters
        ----------
        i : int
            New index for parameter copies. If `i-1` exceeds the number of
            existing copies, then new copies will be added until `i` is a
            valid index.
        new_index : str or None, default='initial'
            Determines how new values are generated if `i` is out of range.
            If `'sample'`   : sample from priors.
            Elif `'mean'`   : mean of priors.
            Elif `'initial'`: initial value of each parameter.
            Elif `'copy'`   : copy of current values.
            Elif `None`     : raise IndexError instead of adding new vectors.

        See also
        --------
        nems.layers.base.Layer.set_index

        """
        for layer in self.layers:
            layer.set_index(index, new_index=new_index)

    def freeze_layers(self, *layer_keys):
        """Invoke `Layer.freeze_parameters()` for each keyed layer.
        
        See also
        --------
        nems.layers.base.Layer.freeze_parameters

        """
        for layer in self.layers.get(*layer_keys):
            layer.freeze_parameters()

    def unfreeze_layers(self, *layer_keys):
        """Invoke `Layer.unfreeze_parameters()` for each keyed layer.
        
        See also
        --------
        nems.layers.base.Layer.unfreeze_parameters

        """
        for layer in self.layers.get(*layer_keys):
            layer.unfreeze_parameters()

    def plot(self, input, **kwargs):
        """Alias for `nems.visualization.model.plot_model`.
        
        By default, the result of each `Layer.evaluate` will be shown.
        
        """
        return plot_model(self, input, **kwargs)

    # added .summary() to mirror tensorflow models, for intuitive comparisons.
    def summary(self):
        """Prints long-form model description (alias for `print(Model)`)."""
        print(self)

    def __str__(self):
        header, tilde_break  = self._repr_helper()
        string = header
        string += tilde_break
        string += ".layers:\n\n"
        # Extra equal_break above first layer, so that its heading looks the
        # same as subsequent layers.
        string += "="*32 + "\n"
        for i, layer in enumerate(self.layers):
            if i != 0:
                # Add blank line between layers if more than one
                string += '\n'
            string += str(layer)
        string += "\n" + tilde_break

        return string

    def __repr__(self):
        header, tilde_break = self._repr_helper()
        string = header
        string += tilde_break
        for i, layer in enumerate(self.layers):
            if i != 0:
                string += "\n\n" # break between layers
            string += layer.__repr__()
        string += "\n" + tilde_break

        return string

    def _repr_helper(self):
        # Get important args/kwargs and string-format as call to constructor.
        # (also attrs, TODO)
        args = []    # TODO  --  what should be here?
        kwargs = {}  # TODO  --  what should be here?
        args_string = ", ".join([f"{a}" for a in args])
        kwargs_string = ", ".join([f"{k}={v}" for k, v in kwargs.items()])
        header = f"{type(self).__name__}({args_string}{kwargs_string})\n"
        tilde_break = "~"*64 + "\n"

        return header, tilde_break


    @classmethod
    def from_keywords(cls, *keywords):
        """Construct a Model from a list of keywords or a model string.
        
        Parameters
        ----------
        keywords : N-tuple of strings
            If first string contains hyphens, it will be interpreted as a
            "model string" where each hyphen separates two keywords.

        Returns
        -------
        Model

        See also
        --------
        nems.layers.base.Layer.from_keyword
        nems.scripts.keywords
        
        """
        # Check for kw1-kw2-... (mode; string) format.
        # If so, split into list of keywords
        split = keywords[0].split('-')
        if len(split) > 1:
            keywords = split
        # Get Layer instances by invoking `Layer.from_keyword` through registry.
        layers = [keyword_lib[kw] for kw in keywords]
        return cls(layers=layers)

    # Add compatibility for saving to json
    def to_json(self):
        """Encode a Model as a dictionary.

        TODO: after specifying some built-in Models (e.g. subclasses), determine
              if `Model.<to/from>_json` need to be updated to support those.
              As long as they're just adding specific Layers the base versions
              should work, but not sure exactly how that's going to work yet.

        Returns
        -------
        data : dict

        See also
        --------
        `nems.tools.json`

        """

        data = {
            'layers': list(self._layers.values()),
            'name': self.name,
            'dtype': self.dtype.__name__,
            'meta': self.meta,
            'results': self.results
        }

        return data

    @classmethod
    def from_json(cls, json):
        """Decode a Model from a dictionary.

        Returns
        -------
        Model

        See also
        --------
        `nems.tools.json`

        """
        model = cls(layers=json['layers'], name=json['name'], 
                    dtype=getattr(np, json['dtype']), meta=json['meta'])
        model.results = json['results']
        return model

    def copy(self):
        """Returns a deep copy of Model without a backend or fit results.
        
        Notes
        -----
        FitResults and Backend objects are removed because they may contain
        state-dependent objects from other packages that cannot copy correctly.
        
        """
        backend = self.backend
        self.backend = None
        copied_model = copy.deepcopy(self)
        self.backend = backend
    
        return copied_model

    def __eq__(self, other):
        if isinstance(other, Model):
            # Layer names don't need to match, but their content does.
            return all([one == two for one, two in
                        zip(self._layers.values(), other._layers.values())])
        else:
            return NotImplemented


class _LayerDict:
    """Wrapper for Layer._layers to enable fancy-indexed gets.

    Supports: integer and string indexing, multi-indexing (one type at a time).
    Note that index assignment is not supported. To change a Model's Layers,
    use `Model.add_layers`, `Model.remove_layers`, etc.

    Examples
    --------
    >>> layers = _LayerDict({'one': 1, 'two': 2, 'three': 3})
    >>> layers
    {'one': 1, 'two': 2, 'three': 3}
    >>> layers[0]
    1
    >>> layers['one']
    1
    >>> layers['one', 'three']
    1, 3
    >>> layers['one', 0]
    KeyError: 0
    >>> layers['two'] = 22
    TypeError: '_LayerDict' object does not support item assignment

    """
    def __init__(self, _dict):
        self._dict = _dict
        self._values = list(_dict.values())

    def __getitem__(self, keys):
        # tuple([]) wrapper to enable multiple keys with Model.layers[] syntax.
        if isinstance(keys, (str, int, slice)):
            keys = tuple([keys])
        value = self.get(*keys, default=None)

        # Raise KeyError if any keys returned None
        if value is None:
            raise KeyError(keys)
        elif isinstance(value, list):
            none_in_list = [x is None for x in value]
            if np.any(none_in_list):
                raise KeyError(keys)

        return value

    def get(self, *keys, default=None):
        # no keys, get all layers
        if keys == ():
            layers = self._values
        elif isinstance(keys[0], slice):
            layers = self._values[keys[0]]
        else:
            # Require all keys str or all int, mixing not allowed.
            # This is meant to discourage hard-to-read code.
            if isinstance(keys[0], int):
                container = self._values
            else:
                container = self._dict
            layers = []
            for key in keys:
                try:
                    layer = container[key]
                except (IndexError, KeyError):
                    layer = default
                layers.append(layer)
        
        # List wrapper (to replace tuple) is just for output consistency should
        # be no practical difference in most cases.
        # Unwrap instead if it's a singleton list, *unless* keys was slice.
        if isinstance(layers, (tuple, list)):
            if (len(layers) == 1) and not isinstance(keys[0], slice):
                layers = layers[0]
            elif isinstance(layers, tuple):
                layers = list(layers)

        return layers

    def __iter__(self):
        """Iterate over Layers (not keys)."""
        return iter(self._values)

    def __len__(self):
        return len(self._values)

    def keys(self):
        return self._dict.keys()

    def items(self):
        return self._dict.items()

    def values(self):
        return self._values

    def __repr__(self):
        return self._dict.__repr__()



from scipy import stats
import numpy as np

from .base import Distribution


class Normal(Distribution):
    """Normal prior.

    Parameters
    ----------
    mean : scalar or ndarray
        Mean of distribution
    sd : scalar or ndarray
        Standard deviation of distribution

    Examples
    --------
    Define a scalar prior for a single coefficient
    >>> weights = Normal(mean=3, sd=5)
    >>> weights.mean()
    3
    Define an array of priors with different means but same standard
    deviation
    >>> weights = Normal(mean=[1, 5], sd=[1, 1])
    >>> weights.mean()
    [1, 5]
    Define an array of priors with same mean and standard deviation
    >>> weights = Normal(mean=[3, 3], sd=[1, 1])
    >>> weights.mean()
    [3, 3]
    
    """

    def __init__(self, mean, sd):
        self._mean = np.asarray(mean)
        self._sd = np.asarray(sd)
        self.distribution = stats.norm(loc=self._mean, scale=self._sd)

    def __repr__(self):
        mean = self.value_to_string(self._mean)
        sd = self.value_to_string(self._sd)
        return 'Normal(μ={}, σ={})'.format(mean, sd)



"""Tools for looking up python objects from strings."""
from functools import partial


class FindCallable:

    def __init__(self, _dict, header='Function', ignore_case=True):
        """Matches a name to a callable object within one module or package.
        
        Case-insensitive.

        Parameters
        ----------
        _dict : dict.
            Keys are names of callables, values are the callables.
        header : str; default='Function'.
            Leading string for error messages. By default, messages start with
            "Function name '{name}' ..."
        ignore_case : bool; default=True.
            If False, match case of dictionary keys when looking up names.
        
        """
        self._dict = _dict
        if ignore_case:
            self._lookup = partial(case_insensitive_lookup, return_none=True)
        else:
            self._lookup = lambda d, k: d.get(k, None)

        self.not_found = f"{header} name " + "'{name}' could not be found."
        self.not_callable = f"{header} name " + "'{name}' is not callable." 

    def __call__(self, name):
        """Get a callable corresponding to `name`.
        
        If `name` is a key in `nicknames`, then `nicknames[name]` will be
        returned. Otherwise `globals()[name]` will be returned.

        Parameters
        ----------
        name : str

        Returns
        -------
        callable

        Raises
        ------
        TypeError, if object to be returned is either `None` or not callable.
        
        """
        fn = self._lookup(self._dict, name)
        if fn is None:
            raise TypeError(self.not_found.format(name=name))
        if not callable(fn):
            raise TypeError(self.not_callable.format(name=name))

        return fn

def case_insensitive_lookup(_dict, key, return_none=False):
    """Look for `key` in `_dict` but ignore case.
    
    Parameters
    ----------
    _dict : dict.
        Dictionary with string keys.
    key : str.
    return_none : bool; default=False.
        If True, and `key` is not found in `_dict`, return None instead of
        raising a KeyError.

    Returns
    -------
    _dict[key] or None
        See `return_none` parameter.

    Raises
    ------
    KeyError
        If `key` is not in `_dict` and `return_none is False`.
    
    """
    keys = list(_dict.keys())
    lowercase_keys = [k.lower() for k in keys]
    unique_lowercase = list(set(lowercase_keys))
    if len(unique_lowercase) < len(lowercase_keys):
        # This would return the first matching index, which could cause
        # confusion. Instead, make the user aware of the issue.
        # Example: don't have two functions named `MSE` and `mse`.
        raise ValueError(
            "Dictionary contains keys that only differ by case, cannot "
            "perform a case-insensitive lookup."
            )

    try:
        idx = lowercase_keys.index(key.lower())
        v = _dict[keys[idx]]
    except ValueError:
        # key not in the list
        if return_none:
            v = None
        else:
            raise KeyError('key')

    return v



"""Collection of plotting utilities for NEMS objects.

Contents
--------
    fitter.py : Plot error vs fit iteration, prediction vs target, etc.
    layers.py : Layer-specific visualizations, like `stp_input_output`.
    model.py  : Whole-model plots, like `plot_model`.
    tools.py  : Helper functions for formatting matplotlib objects.

"""

from .model import plot_model, plot_layer, simple_strf, plot_strf, input_heatmap
from .layers import stp_input_output
from .fitter import (
    prediction_vs_target, iteration_vs_error, evals_per_iteration,
    parameter_space_pca, parameter_pca_table, scipy_gradient
    )
from .tools import standardize_axes



import collections.abc
import math
import copy

import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from .tools import ax_remove_box, ax_bins_to_seconds


_DEFAULT_PLOT_OPTIONS = {
    'skip_plot_options': False,
    'show_x': False, 'xlabel': None, 'xmax_scale': 1, 'xmin_scale': 1,
    'show_y': True, 'ylabel': None, 'ymax_scale': 1, 'ymin_scale': 1,
    'show_seconds': True,
    'legend': False,
    # Right of axis by default, aligned to top
    'legend_kwargs': {
        'frameon': False, 'bbox_to_anchor': (1, 1), 'loc': 'upper left'
        },
    }

def set_plot_options(ax, layer_options, time_kwargs=None):
    """Adjust matplotlib axes object in-place according to `layer_options`.
    
    Parameters
    ----------
    ax : matplotlib.axes.Axes
    layer_options : dict
        Layer-specific options with the same dictionary structure as
        `_DEFAULT_PLOT_OPTIONS`. Layer-specific options take precedence.
    time_kwargs : dict; optional.
        Keyword arguments for `.tools.ax_bins_to_seconds`.

    Returns
    -------
    ops : dict
        Merged plot options from `_DEFAULT_PLOT_OPTIONS` and `layer_options`.

    See also
    --------
    nems.layers.base.Layer.plot_options
    
    """
    _dict = copy.deepcopy(_DEFAULT_PLOT_OPTIONS)
    ops = _nested_update(_dict, layer_options)
    if ops['skip_plot_options']: return

    # x-axis
    ax.xaxis.set_visible(ops['show_x'])
    ax.set_xlabel(ops['xlabel'])
    xmin, xmax = ax.get_xlim()
    ax.set_xlim(xmin*ops['xmin_scale'], xmax*ops['xmax_scale'])

    # y-axis
    ax.yaxis.set_visible(ops['show_y'])
    ax.set_ylabel(ops['ylabel'])
    ymin, ymax = ax.get_ylim()
    ax.set_ylim(ymin*ops['ymin_scale'], ymax*ops['ymax_scale'])

    # Convert bins -> seconds
    if (time_kwargs is not None) and (ops['show_seconds']):
        ax_bins_to_seconds(ax, **time_kwargs)

    # Add legend
    if ops['legend']:
        ax.legend(**ops['legend_kwargs'])
    
    # Remove top and right segments of border around axes
    ax_remove_box(ax)

    return ops


def _nested_update(d, u):
    """Merge two dictionaries that may themselves contain nested dictionaries.
    
    Internal for `set_plot_options`. Using this so that `Layer.plot_options`
    can update some keys of nested dicts (like 'legend_kwargs') without losing
    defaults for other keys.
    TODO: maybe move this to generic NEMS tools? Could be useful elsewhere.
    
    References
    ----------
    https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth

    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = _nested_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def plot_model(model, input, target=None, target_name=None, n=None,
               select_layers=None, n_columns=1, show_titles=True,
               figure_kwargs=None, sampling_rate=None, time_axis='x',
               conversion_factor=1, decimals=2,
               **eval_kwargs):
    """Plot result of `Layer.evaluate` for each Layer in a Model.
    
    Aliased as `Model.plot_output()`.
    TODO: Insert plot of input on first axis, as with old NEMS? But `input` may
          not always be a single spectrogram now, so user would have to specify
          which array to plot, whether to use imshow vs plot vs something else,
          etc... may not be worth it. Alternatively, could add option to insert
          subfigures (possibly at arbitrary positions?) with user-generated
          content, but aligning the x-axis for those could be tricky.

    TODO: SVD's version (plot_model_with_parameters) does this, can just merge
          that logic possibly.

    Parameters
    ----------
    model : Model
        See `nems.models.base.Model`.
    input : ndarray or dict.
        Input data for Model. See `Model.evaluate` for expected format.
    target : ndarray or list of ndarray; optional.
        Target data for Model. See `Model.fit` for expected format. If provided,
        target(s) will be shown on the last axis of the final layer's figure.
    n : int or None; optional.
        Number of Layers to plot. Defaults to all.
    select_layers : str, int, slice, or list; optional.
        Selects layers to plot. Any valid `index` for `Model.layers[index]` is
        acceptable. Ex: `layers = slice(1,3)` to only plot the Layers at indices
        1 and 2, or `layers = ['fir', 'dexp']` to only plot the Layers named
        'fir' and 'dexp'.
        NOTE: This option supercedes `n` if neither is None.
    n_columns : int; default=1.
        Number of columns to arrange plots on (all columns within a row will
        be filled before moving to the next row). Note that values > 1 will
        reduce the width of subplots slightly to make room for the default
        legend placement. However, it may still be necessary to increase
        figure size if the number of columns is large.
    show_titles : bool; default=True.
        Specify whether to show `Layer.name` as a title above each Layer's
        subfigure.
    figure_kwargs : dict or None; optional.
        Keyword arguments for `matplotlib.pyplot.figure`.
        Ex: `figure_kwargs={'figsize': (10,10)}`.
    sampling_rate : float; optional.
        Sampling rate (in Hz) for `input` (and `target` if given), used to
        convert time_bin labels to seconds (if not None).
    time_axis : str; default='x'.
        'x' or 'y', the axis on the resulting Matplotlib axes objects that
        represents time.
    conversion_factor : int; default=1.
        Multiply seconds by this number to get different units. Floating point
        values must correspond to rational numbers. Scientific notation may only
        be used for multiples of 10, and will be converted to an integer.
        Ex: `conversion_factor=1000` to get units of milliseconds.
            `conversion_factor=1/60` to get units of minutes.
    decimals : int; default=2.
        Number of decimal places to show on new tick labels.
    eval_kwargs : dict; optional.
        Additional keyword arguments to supply to `Model.evaluate`.
        Ex: `input_name='stimulus'`.

    Returns
    -------
    fig, axes : (matplotlib.figure.Figure, matplotlib.axes.Axes)
    
    See also
    --------
    nems.models.base.Model
    nems.layers.base.Layer.plot
    nems.visualization.tools.ax_bins_to_seconds

    """

    batch_size = eval_kwargs.get('batch_size', 0)
    if batch_size != 0:
        raise NotImplementedError(
            "Model visualizations are not yet implemented for batched data.\n"
            "To visualize input, collapse sample dimension."
            )

    # Collect options for `ax_bins_to_seconds`
    if sampling_rate is not None:
        time_kwargs = {
            'sampling_rate': sampling_rate, 'time_axis': time_axis,
            'conversion_factor': conversion_factor, 'decimals': decimals
            }
    else:
        time_kwargs = None

    # Determine which figures to plot and set figure layout.
    # One subfigure per Layer, all in a single column by default
    # (otherwise fill columns before rows).
    figure_kwargs = {} if figure_kwargs is None else figure_kwargs
    figure = plt.figure(**figure_kwargs)
    if select_layers is not None:
        layers = model.layers.__getitem__(select_layers)
        if not isinstance(layers, list):
            layers = [layers]
    else:
        layers = model.layers[:n]

    n_rows = math.ceil(len(layers)/n_columns)
    subfigs = figure.subfigures(n_rows, n_columns)

    if isinstance(subfigs, matplotlib.figure.SubFigure):
        # Only one subfigure
        subfigs = np.array([subfigs])

    layer_info = model.generate_layer_data(input, **eval_kwargs)
    iterator = enumerate(zip(layers, subfigs, layer_info))
    for i, (layer, subfig, info) in iterator:
        output = info['out']
        layer.plot(output, fig=subfig, **layer.plot_kwargs)
        if show_titles:
            # Make room for titles
            subfig.subplots_adjust(top=0.8)
            name = layer.name
            subfig.suptitle(f'({model.get_layer_index(name)}) {name}')
        if n_columns > 1:
            # For multiple columns, shrink width of axes a bit to make room for
            # default legend placement.
            subfig.subplots_adjust(right=0.8)
        for ax in subfig.axes:
            set_plot_options(ax, layer.plot_options, time_kwargs=time_kwargs)

    # Add alternating white/gray shading to make Layers easier to distinguish.
    for shaded_fig in subfigs[checkerboard(subfigs)]:
        shaded_fig.patch.set_facecolor((0, 0, 0, 0.075))  # black, low alpha

    # Final x-axis of the final layer in each column is always visible
    # so that time is visually synchronized for all plots above.
    if len(subfigs.shape) == 1:
        bottom_figs = [subfigs[-1]]
    else:
        bottom_figs = subfigs[-1,:]
    for bottom_fig in bottom_figs:
        for ax in bottom_fig.axes:
            ax.xaxis.set_visible(True)

    # Add plot of target if given, on last axis of last subfig.
    last_ax = subfig.axes[-1]
    if target_name is not None:
        target = input[target_name]
    else:
        target_name = 'Target'
    if target is not None:
        if not isinstance(target, list):
            target = [target]
        for i, y in enumerate(target):
            last_ax.plot(y, label=f'{target_name} {i}', alpha=0.3, color='black')
        last_ax.legend(**_DEFAULT_PLOT_OPTIONS['legend_kwargs'])
        last_ax.autoscale()

    return figure


def plot_model_with_parameters(model, input, target=None, target_name=None, n=None,
               select_layers=None, n_columns=1, show_titles=True,
               figure_kwargs=None, sampling_rate=None, time_axis='x',
               conversion_factor=1, decimals=2, plot_input=True,
               **eval_kwargs):
    """TODO: revise doc.

    TODO: Combine with plot_model and/or break out some of the copy-pasted code
          as separate functions.

    Parameters
    ----------
    model : Model
        See `nems.models.base.Model`.
    input : ndarray or dict.
        Input data for Model. See `Model.evaluate` for expected format.
    target : ndarray or list of ndarray; optional.
        Target data for Model. See `Model.fit` for expected format. If provided,
        target(s) will be shown on the last axis of the final layer's figure.
    n : int or None; optional.
        Number of Layers to plot. Defaults to all.
    select_layers : str, int, slice, or list; optional.
        Selects layers to plot. Any valid `index` for `Model.layers[index]` is
        acceptable. Ex: `layers = slice(1,3)` to only plot the Layers at indices
        1 and 2, or `layers = ['fir', 'dexp']` to only plot the Layers named
        'fir' and 'dexp'.
        NOTE: This option supercedes `n` if neither is None.
    n_columns : int; default=1.
        Number of columns to arrange plots on (all columns within a row will
        be filled before moving to the next row). Note that values > 1 will
        reduce the width of subplots slightly to make room for the default
        legend placement. However, it may still be necessary to increase
        figure size if the number of columns is large.
    show_titles : bool; default=True.
        Specify whether to show `Layer.name` as a title above each Layer's
        subfigure.
    figure_kwargs : dict or None; optional.
        Keyword arguments for `matplotlib.pyplot.figure`.
        Ex: `figure_kwargs={'figsize': (10,10)}`.
    sampling_rate : float; optional.
        Sampling rate (in Hz) for `input` (and `target` if given), used to
        convert time_bin labels to seconds (if not None).
    time_axis : str; default='x'.
        'x' or 'y', the axis on the resulting Matplotlib axes objects that
        represents time.
    conversion_factor : int; default=1.
        Multiply seconds by this number to get different units. Floating point
        values must correspond to rational numbers. Scientific notation may only
        be used for multiples of 10, and will be converted to an integer.
        Ex: `conversion_factor=1000` to get units of milliseconds.
            `conversion_factor=1/60` to get units of minutes.
    decimals : int; default=2.
        Number of decimal places to show on new tick labels.
    eval_kwargs : dict; optional.
        Additional keyword arguments to supply to `Model.evaluate`.
        Ex: `input_name='stimulus'`.

    Returns
    -------
    fig, axes : (matplotlib.figure.Figure, matplotlib.axes.Axes)

    See also
    --------
    nems.models.base.Model
    nems.layers.base.Layer.plot
    nems.visualization.tools.ax_bins_to_seconds

    """

    # Collect options for `ax_bins_to_seconds`
    if sampling_rate is not None:
        time_kwargs = {
            'sampling_rate': sampling_rate, 'time_axis': time_axis,
            'conversion_factor': conversion_factor, 'decimals': decimals
            }
    else:
        time_kwargs = None

    # Determine which figures to plot and set figure layout.
    # One subfigure per Layer, all in a single column by default
    # (otherwise fill columns before rows).
    figure_kwargs = {} if figure_kwargs is None else figure_kwargs
    figure = plt.figure(**figure_kwargs)
    if select_layers is not None:
        layers = model.layers.__getitem__(select_layers)
        if not isinstance(layers, list):
            layers = [layers]
    else:
        layers = model.layers[:n]
    layer_info = model.generate_layer_data(input, **eval_kwargs)
    if plot_input:
        n_rows = len(layers)+1
    else:
        n_rows = len(layers)

    spec = figure.add_gridspec(n_rows, 3)
    subaxes = [figure.add_subplot(spec[n, 1:]) for n in range(n_rows)]
    parmaxes = [figure.add_subplot(spec[n+1, 0]) for n in range(n_rows-1)]

    iterator = enumerate(zip(layers, parmaxes, subaxes[1:], layer_info))
    for i, (layer, pax, ax, info) in iterator:

        k = list(layer.parameters.keys())
        if layer.name=='fir':
            if (i>0) & (layers[i-1].name=='wc'):
                plot_strf(layer, wc_layer=layers[i-1], ax=pax)
            else:
                plot_strf(layer, ax=pax)
        elif 'coefficients' in k:
            if len(layer.coefficients.shape)==2:
                pax.plot(layer.coefficients, lw=0.5)
            else:
                pax.imshow(layer.coefficients[:,0,:],
                           aspect='auto', interpolation='none', origin='lower')
            x_pos = pax.get_xlim()[0]
            y_pos = pax.get_ylim()[1]
            title = f'coefficients'
            pax.text(x_pos, y_pos, title, va='top')
        else:
            pax.set_visible(False)

        output = info['out']
        lkw = layer.plot_kwargs
        lkw['lw'] = '0.5'
        layer.plot_options['legend'] = False
        layer.plot(output, ax=ax, **lkw)

        if show_titles:
            x_pos = ax.get_xlim()[0]
            y_pos = ax.get_ylim()[1]
            name = layer.name
            title = f'({model.get_layer_index(name)}) {name}'
            ax.text(x_pos, y_pos, title, va='top')

        set_plot_options(ax, layer.plot_options, time_kwargs=time_kwargs)

    if plot_input:
        # special case to plot input
        ax = subaxes[0]
        if len(input.shape)>2:
            pass
        elif (len(input.shape)>1) & (input.shape[1]>1):
            ax.imshow(input.T, origin='lower', aspect='auto', interpolation='none')
        else:
            ax.plot(input)
        if show_titles:
            title = 'input'
            x_pos = ax.get_xlim()[0]
            y_pos = ax.get_ylim()[1]
            ax.text(x_pos, y_pos, title, va='top')
        ax.set_xlim(subaxes[-1].get_xlim())
        ax.xaxis.set_visible(False)

    # Final x-axis of the final layer in each column is always visible
    # so that time is visually synchronized for all plots above.

    last_ax = subaxes[-1]
    last_ax.xaxis.set_visible(True)

    # Add plot of target if given, on last axis of last subfig.
    if target_name is not None:
        target = input[target_name]
    else:
        target_name = 'Target'
    if target is not None:
        if not isinstance(target, list):
            target = [target]
        for i, y in enumerate(target):
            last_ax.plot(y, label=f'{target_name} {i}', lw=0.5)
        last_ax.legend(**_DEFAULT_PLOT_OPTIONS['legend_kwargs'])
        last_ax.autoscale()
        cc=np.corrcoef(target[0][:,0], output[:,0])[0,1]
    else:
        cc = model.meta.get('r_test',[0])[0]

    figure.suptitle(f"{model.name} cc={cc}", fontsize=10)

    return figure


def simple_strf(model, fir_idx=1, wc_idx=0, ax=None, fig=None):
    """Wrapper for `plot_strf`, gets FIR and WeightChannels from Model.
    
    Parameters
    ----------
    model : nems.models.base.Model
        Model containing a FiniteImpulseResponse Layer and optionally a
        WeightChannels Layer.
    fir_idx : int; default=1.
        Integer index of the FiniteImpulseResponse Layer.
    wc_idx : int; defualt=0.
        Integer index of the WeightChannels Layer, if present.
    ax : Matplotlib axes; optional.
        Axis on which to generate the plot.
    fig : Matplotlib Figure; optional.
        Figure on which to generate the plot.

    Returns
    -------
    Matplotlib Figure
    
    """
    fir_layer, wc_layer = model.layers[fir_idx, wc_idx]
    fig = plot_strf(fir_layer, wc_layer, ax=ax, fig=fig)
    return fig


def plot_strf(fir_layer, wc_layer=None, ax=None, fig=None):
    """Generate a heatmap representing a Spectrotemporal Receptive Field (STRF).
    
    Parameters
    ----------
    fir_layer : nems.layers.FiniteImpulseResponse
        `fir_layer.coefficients.T` will be used to specify the STRF.
    wc_layer : nems.layers.WeightChannels; optional.
        If given, `wc_layer.coefficients @ fir_layer.coefficients.T` will be
        used to specify the STRF (low-rank representation).
    ax : Matplotlib axes; optional.
        Axis on which to generate the plot.
    fig : Matplotlib Figure; optional.
        Figure on which to generate the plot.

    Returns
    -------
    Matplotlib Figure
    
    """
    if ax is not None:
        fig = ax.figure
    else:
        if fig is None:
            fig = plt.figure()
        ax = fig.subplots(1, 1)
    
    if wc_layer is None:
        fir = fir_layer.coefficients
        if len(fir.shape)>2:
            fir=fir[:,:,0]
        strf = fir.T
    else:
        wc = wc_layer.coefficients
        fir = fir_layer.coefficients
        if len(fir.shape)>2:
            wc=wc[:,:,0]
            fir=fir[:,:,0]
        strf = wc @ fir.T

    ax.imshow(strf, aspect='auto', interpolation='none', origin='lower')

    return fig


def plot_layer(output, fig=None, ax=None, **plot_kwargs):
    """Default Layer plot, displays all outputs on a single 2D line plot.
    
    Parameters
    ----------
    output : ndarray or list of ndarray
        Return value of `Layer.evaluate`.
    fig : matplotlib.pyplot.figure.Figure; optional.
        Matplotlib Figure to render the plot on. If not provided, a new figure
        will be generated with default options.
    plot_kwargs : dict
        Additional keyword arguments for `matplotlib.axes.Axes.plot`.

    Returns
    -------
    matplotlib.figure.Figure

    See also
    --------
    .plot_model

    """
    if ax is not None:
        fig = ax.figure
    else:
        if fig is None:
            fig = plt.figure()
        ax = fig.subplots(1, 1)

    all_outputs = [output[...,i] for i in range(output.shape[-1])]
    if all_outputs[0].ndim > 2:
        # TODO: Do something more useful here.
        print("Too many dimensions to plot")
        return fig
    else:
        for output in all_outputs:
            ax.plot(output, **plot_kwargs)

    return fig


def input_heatmap(input, ax=None, extent=None, title='Input',
                  xlabel='Time (bins)', ylabel='Channel', add_colorbar=True):
    """Plot heatmap of `input` with channels increasing from bottom to top.
    
    Parameters
    ----------
    input : np.ndarray, dict, or DataSet.
        Input to `model`. See `nems.Model.evaluate`.
    ax : Matplotlib.axes.Axes; optional.
        Axes on which to plot.
    title : str; optional.
    xlabel : str; default='Time (bins)'.
    ylabel : str; default='Firing Rate (Hz)'.

    Returns
    -------
    matplotlib.axes.Axes
    
    """

    if ax is None: ax = plt.gca()
    im = ax.imshow(input.T, aspect='auto', cmap='binary', interpolation='none',
                   origin='lower', extent=extent)
    if add_colorbar:
        plt.colorbar(im, ax=ax, label='Intensity')
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    return ax


def checkerboard(array):
    """Get checkerboard-spaced indices for a numpy array.

    Using this to keep background-shading of subfigures alternately-spaced.

    Parameters
    ----------
    array : ndarray
    
    Returns
    -------
    indices : boolean ndarray
        Same shape as array, with True and False in a checkerboard pattern.

    References
    ----------
    From user Eelco Hoogendoorn
    https://stackoverflow.com/questions/2169478/how-to-make-a-checkerboard-in-numpy

    Examples
    --------
    >>> x = np.ones(shape=(3,3))
    >>> checkerboard(x)
    array([[False,  True, False],
           [ True, False,  True],
           [False,  True, False]])
    >>> x[checkerboard(x)] = 2
    array([[1., 2., 1.],
           [2., 1., 2.],
           [1., 2., 1.]])

    """
    shape = array.shape
    indices = (np.indices(shape).sum(axis=0) % 2).astype(bool)
    return indices



import numpy as np

from .parameter import Parameter


# TODO: add examples and tests
#       (for both Phi and Parameter)
# TODO: TF compatibility functions? during fitting, values would have to exist
#       as TF primitives, but need to be able to translate to/from
#       TF and NEMS representations.

class Phi:
    """Stores, and manages updates to, Parameters for one Layer."""

    def __init__(self, *parameters):
        """Stores, and manages updates to, Parameters for one Layer.

        In general, Phi instances should not need to be interacted with directly
        unless implementing a new Layer subclass or a related function. Instead,
        parameters should be accessed through Model- or Layer-level methods.

        Additionally, the set of Parameters assigned to a Phi object is meant
        to be fixed at construction. To get a Phi object with some parameters
        either added or removed, construct a new Phi instance or use
        `Phi.modified_copy`.

        Parameters
        ----------
        parameters : N-tuple of Parameter; optional

        See also
        --------
        nems.models.base.Model
        nems.layers.base.Layer
        nems.layers.base.Parameter
        
        """
        self._array = [[]]
        self._index = 0
        self._dict = {}
        self.size = 0
        # Add parameter values to nested list
        for p in parameters:
            self._add_parameter(p)
        # Convert to ndarray
        self._array = np.array(self._array)
        # Cached indexing into `_array` for `get_vector` and `set_vector`.
        self._vector_mask = None
        self._update_vector_mask()

    @property
    def dtype(self):
        return self._array.dtype

    def set_dtype(self, dtype):
        """Change dtype of all Parameter values in-place."""
        self._array = self._array.astype(dtype)

    # TODO: if it becomes necessary, we can relax this restriction and allow
    #       for adding/removing Parameters. But I can't think of a case where
    #       it would really be needed since Layer parameters aren't meant to
    #       change, and this keeps the implementation of other methods simpler.
    def _add_parameter(self, parameter):
        """Add a new parameter to `Phi._dict` and update `Phi._array`.
        
        Sets `parameter.phi`, `parameter.first_index`, and
        `parameter.last_index` to comply with `Phi.ge_vector()` formatting.
        
        This method should only be invoked during construction, since
        `Phi._array` will be converted to ndarray afterward. This limitation
        is intentional: a Phi object is meant to represent a fixed set of
        model parameters. If different parameters are needed after construction,
        create a new Phi object.
        
        """
        # Start at end of existing vector, track size for next parameter
        parameter.first_index = self.size
        self.size += parameter.size
        parameter.last_index = self.size-1
        # Always only one "row" during construction, so use 0 index.
        self._array[0].extend(parameter.initial_value)
        self._dict[parameter.name] = parameter
        parameter.phi = self

    def _get_mask(self, *index_ranges):
        """Get index mask into current vector within `Phi._array`.

        Parameters
        ----------
        index_ranges : N-tuple of 2-tuples
            First tuple entry = first index, second tuple entry = last index.

        Returns
        -------
        mask : boolean ndarray, shape=Phi._array.shape

        Notes
        -----
        Using `mask` for selection will result in a copy of `Phi._array`.
        Using `mask` for assignment will change values of `Phi._array` itself.
        
        """
        mask = np.full(self._array.shape, False)
        for first, last in index_ranges:
            mask[self._index][first:last+1] = True
        return mask

    def _update_vector_mask(self):
        """Update cached copy of current mask for `Phi.<get/set>_vector`.
        
        This method must be invoked any time there is a change to the indices
        within `Phi._array` to which `Phi.<get/set>_vector` would refer
        (i.e. parameters are frozen/unfrozen, `Phi._index` changes, etc).
        
        """
        parameter_ranges = [
            (p.first_index, p.last_index) for p in self._dict.values()
            if not p.is_frozen
            ]
        self._vector_mask = self._get_mask(*parameter_ranges)

    def get_vector(self, as_list=False):
        """Get a copy of `Phi._array` sliced at `Phi._index`.
        
        Parameters
        ----------
        as_list : bool
            If True, return `vector.tolist()` instead of `vector`.

        Returns
        -------
        vector : ndarray or list
        
        """
        vector = self._array[self._vector_mask]
        if as_list:
            vector = vector.tolist()
        return vector

    def get_parameter_from_index(self, i):
        """Get reference to Parameter corresponding to vector index."""
        j = 0
        for p in self._dict.values():
            j += p.size
            if i >= j:
                # Not part of this Parameter
                continue
            else:
                # Part of this Parameter
                return p

    def get_bounds_vector(self, none_for_inf=True):
        """Return a list of bounds from each parameter in `Phi._dict`.
        
        Will not include bounds for frozen parameters.

        Parameters
        ----------
        none_for_inf : bool, default=True
            If True, replace (+/-)`np.inf` with None for compatibility with
            `scipy.optimize.minimize`.
        
        Returns
        -------
        bounds : list of 2-tuples
        
        """
        # Flatten the list returned by each Parameter
        bounds = [b for p in self._dict.values() if not p.is_frozen
                  for b in p.get_bounds_vector(none_for_inf=none_for_inf)]
        return bounds

    @property
    def bounds(self):
        """Get all parameter bounds, as a dict with one key per Parameter."""
        return {k: v.bounds for k, v in self._dict.items()}

    def set_bounds(self, *dct, **kwargs):
        """Set all parameter bounds fromm key-value pairs.
        
        Warnings
        --------
        This can force the current `Parameter.values` out of bounds. To ensure
        valid values, a subsequent `Parameter.sample` or `Parameter.mean` is
        recommended.

        """
        if dct != ():
            _dct = dct[0]
        else:
            _dct = kwargs
        for k, v in _dct.items():
            if v is not None: v = tuple(v)
            self._dict[k].bounds = v

    @property
    def priors(self):
        """Get all parameter priors, as a dict with one key per Parameter."""
        return {k: v.prior for k, v in self._dict.items()}
    
    def set_priors(self, *dct, **kwargs):
        """Set all parameter bounds fromm key-value pairs.
        
        Warnings
        --------
        This can force the current `Parameter.values` to be outside the range
        of `Parameter.prior`. To ensure valid values, a subsequent
        `Parameter.sample` or `Parameter.mean` is recommended.

        """
        if dct != ():
            _dct = dct[0]
        else:
            _dct = kwargs
        for k, v in _dct.items():
            self._dict[k].priors = v

    def get_indices_outof_range(self, vector, as_bool=True):
        """Get indices where `vector < bounds[0]` or `vector > bounds[1]`."""
        bounds = self.get_bounds_vector(none_for_inf=False)
        lower = np.array([b[0] for b in bounds])
        upper = np.array([b[1] for b in bounds])

        if as_bool:
            indices = np.logical_or(vector < lower, vector > upper)
        else:
            check_low = np.argwhere(vector < lower)
            check_high = np.argwhere(vector > upper)
            indices = np.vstack([check_low, check_high]).flatten()
        
        return indices

    def within_bounds(self, vector):
        """False if anywhere `vector < bounds[0]` or `vector > bounds[1]`."""
        return not np.any(self.get_indices_outof_range(vector, as_bool=True))

    def set_vector(self, vector, ignore_checks=False):
        """Set values of `Phi._array` sliced at `Phi._index` to a new vector.

        Parameters
        ----------
        vector : ndarray or list
            New parameter values. Size must match `Phi.get_vector`.
        ignore_checks : bool
            If True, set new values without checking size or bounds.
            (intended as a minor optimization for the scipy fitter).
        
        """
        if not ignore_checks:
            if np.array(vector).size != self.unfrozen_size:
                raise ValueError(f"Size of new vector != Phi.get_vector.")
            if not self.within_bounds(vector):
                bad_indices = self.get_indices_outof_range(vector, as_bool=False)
                raise ValueError("Vector out of bounds at indices:\n"
                                 f"{bad_indices}.")

        self._array[self._vector_mask] = vector

    def _get_parameter_mask(self, p):
        """Get an index mask as in `Phi._get_mask`, but for one Parameter."""
        return self._get_mask((p.first_index, p.last_index))

    def _get_parameter_vector(self, p):
        """Get a sliced copy of `Phi._array` corresponding to one Parameter."""
        mask = self._get_parameter_mask(p)
        return self._array[mask]
        
    def freeze_parameters(self, *parameter_keys):
        """Use parameter values for evaluation only, do not optimize.

        Updates `Phi._vector_mask`.

        Parameters
        ----------
        parameter_keys : N-tuple of str
            Each key must match the name of a Parameter in `Phi._dict`.
            If no keys are specified, all parameters will be frozen.
        
        See also
        --------
        Layer.freeze_parameters
        Parameter.freeze
        
        """
        if parameter_keys == ():
            # no keys given, freeze all parameters
            parameter_keys = list(self._dict.keys())
        for pk in parameter_keys:
            p = self._dict[pk]
            if not p.is_frozen:
                p.freeze()            
        self._update_vector_mask()

    def unfreeze_parameters(self, *parameter_keys):
        """Make parameter values optimizable.

        Updates `Phi._vector_mask`.

        Parameters
        ----------
        parameter_keys : N-tuple of str
            Each key must match the name of a Parameter in `Phi._dict`.
            If no keys are specified, all parameters will be unfrozen.
        
        See also
        --------
        Layer.unfreeze_parameters
        Parameter.unfreeze

        """
        if parameter_keys == ():
            # no keys given, freeze all parameters
            parameter_keys = list(self._dict.keys())
        for pk in parameter_keys:
            p = self._dict[pk]
            if p.is_frozen:
                p.unfreeze()
        self._update_vector_mask()

    @property
    def unfrozen_size(self):
        """Get number of values corresponding to unfrozen parameters."""
        return len(self.get_vector(as_list=True))

    def set_permanent_values(self, *dct, **kwargs):
        """Set parameters to fixed values. The parameters will not unfreeze."""
        if dct != ():
            _dct = dct[0]
        else:
            _dct = kwargs
        for k, v in _dct.items():
            p = self._dict[k]
            p.update(v)
            p.make_permanent()
        
        self._update_vector_mask()

    def sample(self, inplace=False, as_vector=True):
        """Get or set new parameter values by sampling from priors.
        
        Parameters
        ----------
        inplace : bool, default=False
            If True, sampled values will be used to update each Parameter
            (and, in turn, `Phi._array`) inplace. Otherwise, the sampled values
            will be returned without changing current values.
        as_vector : bool, default=True
            If True, return sampled values as a flattened vector instead of a
            list of arrays.

        Return
        ------
        samples : ndarray or list of ndarray

        """
        samples = [p.sample(inplace=inplace) for p in self._dict.values()]
        if not inplace:
            if as_vector:
                unravelled = [np.ravel(s) for s in samples]
                samples = np.concatenate(unravelled)
            return samples

    def mean(self, inplace=False, as_vector=True):
        """Get, or set parameter values to, mean of priors.
        
        Parameters
        ----------
        inplace : bool, default=False
            If True, mean values will be used to update each Parameter
            (and, in turn, `Phi._array`) inplace. Otherwise, means will be
            returned without changing current values.
        as_vector : bool, default=True
            If True, return means as a flattened list instead of a
            list of arrays.

        Return
        ------
        means : list

        """
        means = [p.mean(inplace=inplace) for p in self._dict.values()]
        if not inplace:
            if as_vector:
                unravelled = [np.ravel(m) for m in means]
                means = np.concatenate(unravelled)
            return means


    def set_index(self, i, new_index='initial'):
        """Change which vector to reference within `Phi._array`.

        Updates `Phi._vector_mask`.

        Parameters
        ----------
        i : int
            New index for `Phi._array`. If `i >= len(Phi._array)`, then new
            vectors will be appended until `Phi._array` is sufficiently large.
        new_index : str or None, default='initial'.
            Determines how new vectors are generated if `i` is out of range.
            If `'sample'`   : invoke `Phi.sample()`.
            Elif `'mean'`   : invoke `Phi.mean()`.
            Elif `'initial'`: set to `[p.initial_value for p in <Parameters>]`.
            Elif `'copy'`   : copy current `Phi.get_vector()`.
            Elif `None`     : raise IndexError instead of adding new vectors.

        """
        array_length = len(self._array)
        if i >= array_length:
            # Array isn't that big yet, so add new vector(s).
            new_indices = range(array_length, i+1)
            if new_index == 'sample':
                new_vectors = [self.sample() for j in new_indices]
            elif new_index == 'mean':
                new_vectors = [self.mean() for j in new_indices]
            elif new_index == 'initial':
                new_vectors = [
                    np.concatenate([p.initial_value
                                    for p in self._dict.values()])
                    for j in new_indices
                    ]
            elif new_index == 'copy':
                new_vectors = [self.get_vector() for j in new_indices]
            else:
                # Should be None. Don't add new vectors, raise an error
                # instead. May be useful for testing.
                raise IndexError(f'list index {i} out of range for Phi.')
            # Convert to 2-dim vectors and concatenate after existing vectors
            new_rows = [v[np.newaxis, ...] for v in new_vectors]
            self._array = np.concatenate([self._array] + new_rows)

        self._index = i
        self._update_vector_mask()

    # Provide dict-like interface into Phi._dict
    def __getitem__(self, key):
        return self._dict[key]
    
    def get(self, key, default=None):
        return self._dict.get(key, default)

    def get_values(self, *keys):
        return [self._dict[k].values for k in keys]

    def update(self, *dct, ignore_bounds=False, **kwargs):
        """Update Parameter values (not the Parameters themselves)."""
        if dct != ():
            _dct = dct[0]
        else:
            _dct = kwargs
        for k, v in _dct.items():
            self._dict[k].update(v, ignore_bounds=ignore_bounds)

    def __setitem__(self, key, val):
        """Update Parameter value (not the Parameters itself)."""
        self._dict[key].update(val)

    def __iter__(self):
        return iter(self._dict.values())

    def keys(self):
        return self._dict.keys()

    def items(self):
        return self._dict.items()

    def values(self):
        return self._dict.values()

    def __str__(self):
        footer = f"Index: {self._index}\n" + "-"*16 + "\n"
        string = ""
        for i, p in enumerate(self._dict.values()):
            if i != 0:
                # Add blank line between parameters if more than one
                string += "\n"
            string += str(p)
            string += footer
        string += "\n"

        return string

    def __repr__(self):
        dash_break = "-"*16 + "\n"
        string = ""
        for i, p in enumerate(self._dict.values()):
            if i != 0:
                string += dash_break
            string += p.__repr__()
        return string


    # Add compatibility for saving to .json
    def to_json(self):
        """Encode Phi object as json. See `nems.tools.json`."""
        p = list(self._dict.values())
        frozen_parameters = [k for k, v in self._dict.items() if v.is_frozen]
        data = {
            'args': p,
            'attributes': {
                '_array': self._array,
                '_index': self._index
            },
            'frozen_parameters': frozen_parameters
        }
        return data

    @classmethod
    def from_json(cls, json):
        """Decode Phi object from json. See `nems.tools.json`."""
        phi = cls(*json['args'])
        for k, v in json['attributes'].items():
            setattr(phi, k, v)
        frozen = json['frozen_parameters']
        if len(frozen) > 0:
            phi.freeze_parameters(*json['frozen_parameters'])
        return phi

    def from_dict(dct, default_bounds='infinite'):
        """Construct Phi from a specially formatted dictionary.
        
        Parameters
        ----------
        dct : dict
            Must contain three nested dictionaries at keys 'values', 'prior',
            and 'bounds'. Each dictionary must have the same keys, which will
            be the parameter names. Values within each dictionary will be used
            as arguments for `initial_value`, `prior`, and `bounds`,
            respectively.
        default_bounds : string, default='infinite'
            Determines behavior when `bounds=None`.
            If `'infinite'`  : set bounds to (-np.inf, np.inf)
            If `'percentile'`: set bounds to tails of `Parameter.prior`.
                (prior.percentile(0.0001), prior.percentile(0.9999))

        See also
        --------
        Parameter.__init__
        
        """
        parameters = []
        for name, value in dct['values'].items():
            value = np.array(value)
            prior = dct['prior'][name]
            bounds = dct['bounds'][name]
            p = Parameter(name, shape=value.shape, prior=prior, bounds=bounds,
                          default_bounds=default_bounds, initial_value=value)
            parameters.append(p)
        phi = Phi(*parameters)
        return phi

    def modified_copy(self, keys_to_keep, parameters_to_add):
        """TODO."""
        #       ref `keys_to_keep` to store Parameter objects,
        #       combine with parameters_to_add,
        #       build new phi,
        #       overwrite part of new array with copy of old array
        #       copy old index
        raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, Phi):
            try:
                arrays_match = np.allclose(self._array, other._array)
            except:
                arrays_match = False
            try:
                masks_match = np.allclose(self._vector_mask, other._vector_mask)
            except:
                masks_match = False
            conditions = [
                arrays_match, masks_match,
                (self._dict == other._dict),
                (self._index == other._index)
            ]
            return all(conditions)
        else:
            return NotImplemented





class ShapeError(Exception):
    def __init__(self, layer, **shapes):
        """Raise in Layer.evaluate if data or parameter shapes are misaligned.
        
        Parameters
        ----------
        layer : .layer.Layer
            Layer that the exception was raised from.
        shapes : dict
            Keys indicate names of shapes in the error message, values indicate
            the associated shape.

        See also
        --------
        nems.layers.WeightChannels.evaluate  (example usage)
        
        """
        self.layer = layer
        self.shapes = shapes
    def __str__(self):
        string = (
            f"Shape mismatch in {type(self.layer).__name__}.evaluate.\n")
        for k, v in self.shapes.items():
            string += f"{k} has shape: {v.__repr__()}\n"
        return string



import numpy as np
import numexpr as ne

from nems.registry import layer
from nems.distributions import Normal
from .base import Layer, Phi, Parameter
from .tools import require_shape, pop_shape


class StaticNonlinearity(Layer):
    """Apply a nonlinear transformation to input(s).
    
    TODO: Test if current implementations will work with higher dim data.

    Parameters
    ----------
    shape : N-tuple of int (usually N=1)
        Determines the shape of each Parameter in `.parameters`.
        First dimension should match the last dimension of the input.
        Note that higher-dimesional shapes are also allowed, but overall
        Layer design is intended for 1-dimensional shapes.

    Attributes
    ----------
    skip_nonlinearity : bool
        If True, don't apply `nonlinearity` during `evaluate`. Still apply
        `input += shift`, if `'shift' in StaticNonlinearity.parameters`.
    
    """

    def __init__(self, **kwargs):
        require_shape(self, kwargs, minimum_ndim=1)
        super().__init__(**kwargs)
        self._skip_nonlinearity = False
        self._unfrozen_parameters = []

    def skip_nonlinearity(self):
        """Don't use `nonlinearity`, freeze nonlinear parameters."""
        self._unfrozen_parameters = [p.name for p in self.parameters
                                     if not p.is_frozen]
        self.freeze_parameters()
        self.unfreeze_parameters('shift')
        self._skip_nonlinearity = True

    def unskip_nonlinearity(self):
        """Use `nonlinearity`, unfreeze nonlinear parameters."""
        # Only unfreeze parameters that were previously unfrozen, but then
        # frozen by `skip_nonlinearity`.
        self.unfreeze_parameters(*self._unfrozen_parameters)
        self._unfrozen_parameters = []
        self._skip_nonlinearity = False

    def evaluate(self, input):
        """Apply `nonlinearity` to input(s). This should not be overwriten."""
        if not self._skip_nonlinearity:
            output = self.nonlinearity(input)
        else:
            # TODO: This works for time on 0-axis and 1-dim parameters,
            #       but need to add option to make this more general.
            # If there's a `shift` parameter for the subclassed nonlinearity,
            # still apply that. Otherwise, pass through inputs.
            output = input + self.parameters.get('shift', 0)
        return output

    def nonlinearity(self, input):
        """Pass through input(s). Subclasses should overwrite this."""
        return input

    def as_tensorflow_layer(self, **kwargs):
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        class StaticNonlinearityTF(NemsKerasLayer):
            def call(self, inputs):
                # TODO: why identity?
                return tf.identity(inputs + self.shift)

        return StaticNonlinearityTF(self, **kwargs)


class LevelShift(StaticNonlinearity):
    """Applies a scalar shift to each input channels.
    
    Notes
    -----
    While `LevelShift.evaluate` is linear, this Layer is grouped with
    `StaticNonlinearity` because of its close relation to these other layers.
    In short, we have found in the past that it is often helpful to "turn off"
    a nonlinearity during fitting, but still shift the input.
    
    """

    def initial_parameters(self):
        """Get initial values for `StaticNonlinearity.parameters`.
        
        Layer parameters
        ----------------
        shift : scalar or ndarray
            Value(s) that are added to input(s) prior to rectification. Shape
            (N,) must match N channels per input.
            Prior:  TODO
        
        """
        # TODO: explain choice of priors.
        prior = Normal(
            np.zeros(shape=self.shape), 
            np.ones(shape=self.shape)/100
            )

        shift = Parameter('shift', shape=self.shape, prior=prior)
        return Phi(shift)

    def nonlinearity(self, input):
        """constant shift

        Notes
        -----
        Simply add a constant shift to the signal

        """
        shift, = self.get_parameter_values()
        return input + shift

    @layer('lvl')
    def from_keyword(keyword):
        """Construct LevelShift from a keyword.

        Keyword options
        ---------------
        {digit}x{digit}x ... x{digit} : N-dimensional shape; required.

        Returns
        -------
        LevelShift

        See also
        --------
        Layer.from_keyword

        """
        options = keyword.split('.')
        shape = pop_shape(options)

        return LevelShift(shape=shape)

    @property
    def plot_kwargs(self):
        """Add incremented labels to each output channel for plot legend.

        See also
        --------
        Layer.plot

        """
        kwargs = {
            'label': [f'Channel {i}' for i in range(self.shape[1])]
        }
        return kwargs

    @property
    def plot_options(self):
        """Add legend at right of plot, with default formatting.

        Notes
        -----
        The legend will grow quite large if there are many output channels,
        but for common use cases (< 10) this should not be an issue. If needed,
        increase figsize to accomodate the labels.

        See also
        --------
        Layer.plot

        """
        return {'legend': False}


class DoubleExponential(StaticNonlinearity):
    """TODO: doc here? maybe just copy .evaluate?"""

    def initial_parameters(self):
        """Get initial values for `DoubleExponential.parameters`.
        
        Layer parameters
        ----------------
        base : scalar or ndarray
            Y-axis height of the center of the sigmoid.
            Shape (N,) must match N input channels (same for other parameters),
            such that one sigmoid transformation is applied to each channel.
            Prior:  Normal(mean=0, sd=1)
            Bounds: TODO
        amplitude : scalar or ndarray
            Y-axis distance from ymax asymptote to ymin asymptote
            Prior:  Normal(mean=5, sd=1.5)
            Bounds: TODO
        shift : scalar or ndarray
            Centerpoint of the sigmoid along x axis
            Prior:  Normal(mean=0, sd=1)
            Bounds: TODO
        kappa : scalar or ndarray
            Sigmoid curvature. Larger numbers mean steeper slop.
            Prior:  Normal(mean=1, sd=10)
            Bounds: TODO

        Returns
        -------
        nems.layers.base.Phi
        
        """
        # TODO: explain choices for priors.
        zero = np.zeros(shape=self.shape)
        one = np.ones(shape=self.shape)
        phi = Phi(
            Parameter('base', shape=self.shape, prior=Normal(-one, one/5)),
            Parameter('amplitude', shape=self.shape, prior=Normal(2*one, one/5)),
            Parameter('shift', shape=self.shape, prior=Normal(zero, one/5)),
            Parameter('kappa', shape=self.shape, prior=Normal(one, one/5))
            )
        return phi

    def nonlinearity(self, input):
        """Apply sigmoid transform to input x: $b+a*exp[-exp(-exp(k)(x-s)]$.
        
        See Thorson, Liénard, David (2015).
        
        """
        base, amplitude, shift, kappa = self.get_parameter_values()

        if (input.shape[-1] < base.shape[-1]) or (not self._inplace_ok):
            # First condition means output will be larger than input, so we
            # can't store it in the same array.
            out = None
        else:
            out = input

        output = ne.evaluate(
            "base + amplitude*exp(-exp(-exp(kappa)*(input+shift)))",
            out=out
            )

        return output

    @layer('dexp')
    def from_keyword(keyword):
        """Construct DoubleExponential from keyword.

        Keyword options
        ---------------
        {digit}x{digit}x ... x{digit} : N-dimensional shape; required.

        Returns
        -------
        DoubleExponential

        See also
        --------
        Layer.from_keyword
        
        """
        options = keyword.split('.')
        shape = pop_shape(options)
        
        return DoubleExponential(shape=shape)

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        if self._skip_nonlinearity:
            return super().as_tensorflow_layer(**kwargs)
        else:
            class DoubleExponentialTF(NemsKerasLayer):
                def call(self, inputs):
                    exp = tf.math.exp(-tf.math.exp(
                        -tf.math.exp(self.kappa) * (inputs + self.shift)
                        ))
                    return self.base + self.amplitude * exp

            return DoubleExponentialTF(self, **kwargs)


class RectifiedLinear(StaticNonlinearity):
    """TODO: doc here? maybe just copy .evaluate?"""
    def __init__(self, no_shift=True, no_offset=True, no_gain=True, **kwargs):
        super().__init__(**kwargs)
        fixed_parameters = {}
        shift, offset, gain = self.get_parameter_values()
        if no_shift: fixed_parameters['shift'] = np.full_like(shift, 0)
        if no_offset: fixed_parameters['offset'] = np.full_like(offset, 0)
        if no_gain: fixed_parameters['gain'] = np.full_like(gain, 1)
        self.set_permanent_values(**fixed_parameters)

    def initial_parameters(self):
        """Get initial values for `RectifiedLinear.parameters`.
        
        Layer parameters
        ----------------
        shift : scalar or ndarray
            Value(s) that are added to input prior to rectification. Shape
            (N,) must match N channels per input.
            Prior:  Normal(mean=-0.1, sd=1/sqrt(N))
        offset : scalar or ndarray
            Value(s) that are added to input after rectification.
            Prior:  TODO
        gain : scalar or ndarray
            Rectified input(s) will be multiplied by this (i.e. slope of the
            linear portion for each output).
            Prior:  TODO
        
        """
        # TODO: explain choice of prior.
        zero = np.zeros(shape=self.shape)
        one = np.ones(shape=self.shape)
        shift_prior = {'mean': zero+0.05, 'sd': one/100}
        offset_prior = {'mean': zero-0.05, 'sd': one/100}
        gain_prior = {'mean': one, 'sd': one/100}
        phi = Phi(
            Parameter('shift', shape=self.shape, prior=Normal(**shift_prior)),
            Parameter('offset', shape=self.shape, prior=Normal(**offset_prior)),
            Parameter('gain', shape=self.shape, prior=Normal(**gain_prior))
            )

        return phi

    def nonlinearity(self, input):
        """Implements `y = offset + gain * rectify(x - shift)`.
        
        By default, `offset=0, shift=0, gain=1` and this is equivalent to
        standard linear rectification: `y = 0 if x < 0, else x`. 

        Notes
        -----
        The negative of `shift` is used so that its interpretation in
        `StaticNonlinearity.evaluate` is the same as for other subclasses.
        
        """

        shift, offset, gain = self.get_parameter_values()

        if (input.shape[-1] < shift.shape[-1]) or (not self._inplace_ok):
            # First condition means output will be larger than input, so we
            # can't store it in the same array.
            out = None
        else:
            out = input

        output = ne.evaluate('offset + gain*((input + shift)*(input > -shift))')
        return output

    @layer('relu')
    def from_keyword(keyword):
        """Construct RectifiedLinear from a keyword.

        Keyword options
        ---------------
        {digit}x{digit}x ... x{digit} : N-dimensional shape; required.

        Returns
        -------
        RectifiedLinear

        See also
        --------
        Layer.from_keyword

        """
        options = keyword.split('.')
        no_shift = True
        no_offset = True
        no_gain = True
        shape = pop_shape(options)

        for op in options[1:]:
            if op == 's':
                no_shift = False
            elif op == 'o':
                no_offset = False
            elif op == 'g':
                no_gain = False

        relu = RectifiedLinear(
            shape=shape, no_shift=no_shift, no_offset=no_offset,
            no_gain=no_gain
            )

        return relu

    def as_tensorflow_layer(self, **kwargs):
        """TODO: docs"""
        import tensorflow as tf
        from nems.backends.tf import NemsKerasLayer

        if self._skip_nonlinearity:
            return super().as_tensorflow_layer(**kwargs)
        else:
            class RectifiedLinearTF(NemsKerasLayer):
                def call(self, inputs):
                    rectified  = tf.nn.relu(inputs + self.shift)
                    return self.offset + self.gain * rectified

        return RectifiedLinearTF(self, **kwargs)

# Optional alias
class ReLU(RectifiedLinear):
    pass
